import argparse
import os

import numpy as np
from git_commit import get_git_commit

import vamb


def split_clusters_by_sample(clusters):
    nbhds_split = dict()
    for i, (nbhd_id, nbhd_cs) in enumerate(clusters.items()):
        samples_in_nbhd = set([c.split("C")[0] for c in nbhd_cs])

        nbhd_S_d = {S: set() for S in samples_in_nbhd}

        for c in nbhd_cs:
            nbhd_S_d[c.split("C")[0]].add(c)

        for S, cs in nbhd_S_d.items():
            nbhds_split["%s_%s" % (str(nbhd_id), S)] = cs

    return nbhds_split


if __name__ == "__main__":
    # Initialize the parser
    parser = argparse.ArgumentParser(
        description="Classify clusters based upon the contig plasmid, virus, and Organism geNomad scores"
    )
    # Add optional arguments with `--flags`
    parser.add_argument(
        "--clusters",
        required=True,
        help="Clusters generated with the community-based clustering algorithm",
    )
    parser.add_argument(
        "--composition",
        required=True,
        help="Composition object containing contignames and contig lengths",
    )
    parser.add_argument(
        "--scores",
        required=True,
        type=str,
        help="Path to the file that contains the contig scores",
    )
    parser.add_argument(
        "--outp",
        required=True,
        help="Plasmid clusters file",
    )

    parser.add_argument(
        "--dflt_cls",
        required=True,
        help="clusters generated with the default clustering algorithm, where plasmid contigs will be removed",
    )
    parser.add_argument(
        "--split",
        action="store_true",
        help="if set, contig score aggregation is applied per bin",
    )
    parser.add_argument(
        "--thr",
        default=0.7,
        type=float,
        help="Threshold set to classify bins",
    )
    parser.add_argument(
        "--thr_circ",
        default=0.5,
        type=float,
        help="Threshold set to classify circular contigs as plasmids",
    )

    ## Print git commit so we can debug
    commit_hash = get_git_commit(os.path.abspath(__file__))
    print(f"Git commit hash: {commit_hash}")

    # Parse the arguments
    args = parser.parse_args()
    ## Load contignames aand lenghts
    composition = np.load(args.composition, allow_pickle=True)
    contignames = composition["identifiers"]
    contiglengths = composition["lengths"]
    c_len_d = {c: l for c, l in zip(contignames, contiglengths)}

    ## Load contig scores
    c_genomad_d = {}
    for row in np.loadtxt(args.scores, skiprows=1, dtype=object):
        contig, org_scr, pla_scr, vir_scr = row
        c_genomad_d[contig] = {
            "org": float(org_scr),
            "pla": float(pla_scr),
            "vir": float(vir_scr),
        }

    ## Load clusters and define a plasmid, virus, and organism score per cluster
    clusters_arr = np.loadtxt(args.clusters, skiprows=1, dtype=object)
    cl_cs_d = {cl: set() for cl in set(clusters_arr[:, 0])}
    for cl, c in clusters_arr:
        cl_cs_d[cl].add(c)

    if args.split:
        print("Splitting clusters per sample")
        cl_cs_d = split_clusters_by_sample(cl_cs_d)

    cl_genomadscores_d = {}
    thr_circ = min(args.thr_circ, args.thr)
    for cl, cs in cl_cs_d.items():
        if cl.endswith("_circ"):
            assert len(cs) == 1
            c = list(cs)[0]
            if c_genomad_d[c]["pla"] >= thr_circ:
                cl_genomadscores_d[cl] = {
                    "org": 0,
                    "pla": 1,
                    "vir": 0,
                }
            else:
                cl_genomadscores_d[cl] = {
                    "org": c_genomad_d[c]["org"],
                    "pla": c_genomad_d[c]["pla"],
                    "vir": c_genomad_d[c]["vir"],
                }

        else:
            denominator = np.sum([c_len_d[c] for c in cs])
            numerator_org = np.sum([c_genomad_d[c]["org"] * c_len_d[c] for c in cs])
            numerator_pla = np.sum([(c_genomad_d[c]["pla"]) * c_len_d[c] for c in cs])
            numerator_vir = np.sum([(c_genomad_d[c]["vir"]) * c_len_d[c] for c in cs])

            cl_genomadscores_d[cl] = {
                "org": numerator_org / denominator,
                "pla": numerator_pla / denominator,
                "vir": numerator_vir / denominator,
            }

    print("%i contigs with genomad scores" % (len(c_genomad_d.keys())))

    # Precompute the mask array for all contigs that meet the threshold
    mask_organism_cluster_contigs = np.zeros(len(c_len_d.keys()), dtype=bool)
    organism_cluster_contigs = set()

    # Filter clusters that pass the threshold
    below_threshold_clusters = {
        cl: cs
        for cl, cs in cl_cs_d.items()
        if cl_genomadscores_d[cl]["pla"] < float(args.thr)
    }

    # Use numpy's isin function to optimize masking in bulk
    all_contigs = np.array(list(c_len_d.keys()))
    below_threshold_contigs = set(
        c for cs in below_threshold_clusters.values() for c in cs
    )
    mask_organism_cluster_contigs = np.isin(all_contigs, list(below_threshold_contigs))

    # Add contigs to the set
    organism_cluster_contigs.update(below_threshold_contigs)

    # Write out clusters above the threshold to the file
    with open(args.outp, "w") as f_pla:
        f_pla.write("clustername\tcontigname\n")
        for cl, cs in cl_cs_d.items():
            if cl not in below_threshold_clusters:
                for c in cs:
                    f_pla.write(f"{cl}\t{c}\n")

    ## Load dflt clusters and extract plasmid contigs
    dfltclusters_arr = np.loadtxt(args.dflt_cls, skiprows=1, dtype=object)
    dfltcl_cs_d = {cl: set() for cl in set(dfltclusters_arr[:, 0])}
    for cl, c in dfltclusters_arr:
        if c in organism_cluster_contigs:
            dfltcl_cs_d[cl].add(c)

    ## save new clusters
    f_nonplasmid_cls = args.dflt_cls.replace(
        ".tsv",
        (
            "_geNomadplasclustercontigs_extracted_thr_%s_thrcirc_%s.tsv"
            % (str(args.thr), str(thr_circ))
        ),
    )
    with open(f_nonplasmid_cls, "w") as f:
        f.write("clustername\tcontigname\n")
        for cl, cs in dfltcl_cs_d.items():
            for c in cs:
                f.write("%s\t%s\n" % (cl, c))

    ## Write geNomad scores per plasmid candidate cluster
    with open(args.outp.replace(".tsv", "_gN_scores.tsv"), "w") as f:
        f.write(
            "%s\t%s\t%s\t%s\n"
            % ("clustername", "chromosome_score", "plasmid_score", "virus_score")
        )
        for cl in cl_cs_d.keys():
            if cl not in below_threshold_clusters:
                if "circ" not in cl:
                    f.write(
                        "%s\t%s\t%s\t%s\n"
                        % (
                            cl,
                            np.round(cl_genomadscores_d[cl]["org"], 3),
                            np.round(cl_genomadscores_d[cl]["pla"], 3),
                            np.round(cl_genomadscores_d[cl]["vir"], 3),
                        )
                    )
                else:
                    c = cl.replace("_circ", "")
                    f.write(
                        "%s\t%s\t%s\t%s\n"
                        % (
                            cl,
                            np.round(c_genomad_d[c]["org"], 3),
                            np.round(c_genomad_d[c]["pla"], 3),
                            np.round(c_genomad_d[c]["vir"], 3),
                        )
                    )

    ## Write geNomad scores per non-plasmid candidate cluster
    # First get the aggregated scores per cls
    dfltcl_genomadscores_d = dict()
    for cl, cs in dfltcl_cs_d.items():
        if len(cs) == 0:
            continue
        denominator = np.sum([c_len_d[c] for c in cs])
        numerator_org = np.sum([c_genomad_d[c]["org"] * c_len_d[c] for c in cs])
        numerator_pla = np.sum([(c_genomad_d[c]["pla"]) * c_len_d[c] for c in cs])
        numerator_vir = np.sum([(c_genomad_d[c]["vir"]) * c_len_d[c] for c in cs])

        dfltcl_genomadscores_d[cl] = {
            "org": numerator_org / denominator,
            "pla": numerator_pla / denominator,
            "vir": numerator_vir / denominator,
        }

    ## Write scores
    with open(f_nonplasmid_cls.replace(".tsv", "_gN_scores.tsv"), "w") as f:
        f.write(
            "%s\t%s\t%s\t%s\n"
            % ("clustername", "chromosome_score", "plasmid_score", "virus_score")
        )
        for cl in dfltcl_cs_d.keys():
            if len(dfltcl_cs_d[cl]) == 0:
                continue
            if "circ" not in cl:
                f.write(
                    "%s\t%s\t%s\t%s\n"
                    % (
                        cl,
                        np.round(dfltcl_genomadscores_d[cl]["org"], 3),
                        np.round(dfltcl_genomadscores_d[cl]["pla"], 3),
                        np.round(dfltcl_genomadscores_d[cl]["vir"], 3),
                    )
                )
            else:
                c = cl.replace("_circ", "")
                f.write(
                    "%s\t%s\t%s\t%s\n"
                    % (
                        cl,
                        np.round(c_genomad_d[c]["org"], 3),
                        np.round(c_genomad_d[c]["pla"], 3),
                        np.round(c_genomad_d[c]["vir"], 3),
                    )
                )



import argparse
import os
from concurrent.futures import ProcessPoolExecutor, as_completed

import numpy as np
import pysam
from git_commit import get_git_commit


# Function to process a single BAM file
def process_bam(bam_path, max_insert_len):
    """
    This function parses a bam file, and identifies paired-end reads where each read mate maps to the end of the contig, within certain distance

    Args:
        bam_path (str): Path of the bam file containing the paired end reads from a sample mapping to all contigs.
        max_insert_len (in): Max distance between the end of the contig and the read mapping position for it to be considered.

    Returns:
        local_circular_contigs (set): set of contigs with paired-end reads mapping to contig ends.
        contig_reads_d (dict): dictionary containing the reads and read position for circularized contigs.
        allcontig_reads_d (dict): dictionary containing the reads and read position for all contigs.
    """
    local_circular_contigs = set()
    contig_reads_d = dict()
    allcontig_reads_d = dict()

    with pysam.AlignmentFile(bam_path, "rb") as bamfile:
        reference_lengths = {}

        for read in bamfile:
            if (
                read.is_unmapped
                or read.mate_is_unmapped
                or read.next_reference_name != read.reference_name
            ):
                continue
            is_start_mapping = (
                read.is_reverse and read.reference_start < max_insert_len
            ) and (
                not read.mate_is_reverse
                and read.next_reference_start
                > (read.reference_length - read.query_length - max_insert_len)
            )
            is_end_mapping = (
                not read.is_reverse
                and (
                    read.reference_start
                    > read.reference_length - read.query_length - max_insert_len
                )
            ) and (read.mate_is_reverse and read.next_reference_start < max_insert_len)
            if not (is_start_mapping or is_end_mapping):
                continue

            reference_name = read.reference_name
            if reference_name not in reference_lengths:
                reference_lengths[reference_name] = bamfile.get_reference_length(
                    reference_name
                )

            reference_length = reference_lengths[reference_name]

            read_start = read.reference_start + 1
            read_end = read.reference_end

            mate_start = read.next_reference_start + 1
            mate_end = mate_start + read.query_length - 1

            read_dist_5_prime = read_start
            read_dist_3_prime = reference_length - read_end

            mate_dist_5_prime = mate_start
            mate_dist_3_prime = reference_length - mate_end
            if (read_dist_5_prime + mate_dist_3_prime <= max_insert_len) or (
                mate_dist_5_prime + read_dist_3_prime <= max_insert_len
            ):
                local_circular_contigs.add(reference_name)
                if reference_name not in contig_reads_d.keys():
                    contig_reads_d[reference_name] = []
                contig_reads_d[reference_name].append(
                    [
                        read_start,
                        mate_start,
                        not read.is_reverse,
                        read.mate_is_reverse,
                    ]
                )
                if reference_name not in allcontig_reads_d.keys():
                    allcontig_reads_d[reference_name] = []
                allcontig_reads_d[reference_name].append(
                    [
                        read_start,
                        mate_start,
                        not read.is_reverse,
                        read.mate_is_reverse,
                    ]
                )

            else:
                if reference_name not in allcontig_reads_d.keys():
                    allcontig_reads_d[reference_name] = []
                allcontig_reads_d[reference_name].append(
                    [
                        read_start,
                        mate_start,
                        not read.is_reverse,
                        read.mate_is_reverse,
                    ]
                )
    return local_circular_contigs, contig_reads_d, allcontig_reads_d


# Function to process all BAM files in parallel
def process_all_bams_parallel(dir_bams, max_insert_len):
    """Parellelize the parsing of the bam files

    Args:
        dir_bams (str): Path that contains bam files from all samples.
        max_insert_len (int): Max distance between the end of the contig and the read mapping position for it to be considered.

    Returns:
        contig_reads_d: dictionary containing the reads and read position for circularized contigs.
        allcontig_reads_d: dictionary containing the reads and read position for all contigs.
    """
    bam_files = [
        os.path.join(dir_bams, f)
        for f in os.listdir(dir_bams)
        if "bam" in f and ".bai" not in f  # f.endswith("bam")
    ]
    contig_reads_d = dict()
    allcontig_reads_d = dict()

    with ProcessPoolExecutor() as executor:
        futures = {
            executor.submit(process_bam, bam, max_insert_len): bam for bam in bam_files
        }
        for future in as_completed(futures):
            bam = futures[future]
            try:
                (
                    local_circular_contigs,
                    contig_samplereads_d,
                    allcontig_samplereads_d,
                ) = future.result()

                # Update global sets with results from the parallel tasks
                circular_contigs.update(local_circular_contigs)

                # update contig_reads_d
                for c in contig_samplereads_d.keys():
                    if c not in contig_reads_d.keys():
                        contig_reads_d[c] = contig_samplereads_d[c]
                    else:
                        contig_reads_d[c] += contig_samplereads_d[c]

                # update allcontig_reads_d
                for c in allcontig_samplereads_d.keys():
                    if c not in allcontig_reads_d.keys():
                        allcontig_reads_d[c] = allcontig_samplereads_d[c]
                    else:
                        allcontig_reads_d[c] += allcontig_samplereads_d[c]

                print(f"{os.path.basename(bam)} completed")

            except Exception as e:
                print(f"Error processing {bam}: {e}")
                raise e
    return contig_reads_d, allcontig_reads_d


def save_circular_contig_clusters(circular_contigs, cl_f):
    with open(cl_f, "w") as f:
        f.write("%s\t%s\n" % ("clustername", "contigname"))
        for c in circular_contigs:
            f.write("%s_%s\t%s\n" % (c, "circ", c))


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Extract circular contigs")

    # Add optional arguments with `--flags`
    parser.add_argument(
        "--dir_bams",
        required=True,
        type=str,
        help="Path to the directory that contains the bam files.",
    )

    parser.add_argument(
        "--max_insert",
        required=False,
        type=int,
        help="Max length of sequence between paired-end reads [50].",
        default=50,
    )

    parser.add_argument(
        "--outcls",
        type=str,
        required=True,
        help="Path were circular contigs will be dumped.",
    )

    parser.add_argument(
        "--outreads",
        type=str,
        help="Path were contig-read information will be stored for contigs circularizable.",
    )
    parser.add_argument(
        "--outallreads",
        type=str,
        help="Path were contig-read information will be stored for all contigs.",
    )

    # Parse the arguments
    args = parser.parse_args()

    ## Print git commit so we can debug
    commit_hash = get_git_commit(os.path.abspath(__file__))
    print(f"Git commit hash: {commit_hash}")

    max_insert_len = args.max_insert

    plas_recov = set()
    circular_contigs = set()  ## GLOBAL VARIABLE

    # Run the parallel processing function
    contig_reads_d, allcontig_reads_d = process_all_bams_parallel(
        args.dir_bams, max_insert_len
    )
    if args.outcls != None:
        save_circular_contig_clusters(circular_contigs, args.outcls)
    else:
        pass
    if args.outreads != None:
        with open(args.outreads, "w") as f:
            f.write(
                "%s\t%s\t%s\t%s\t%s\n"
                % ("contig", "start", "start_mate", "is_fw", "mate_is_rv")
            )
            for c in contig_reads_d.keys():
                for read_info in contig_reads_d[c]:
                    s, s_m, read_is_fw, mate_is_rv = read_info
                    f.write(
                        "%s\t%i\t%i\t%s\t%s\n"
                        % (c, s, s_m, str(read_is_fw), str(mate_is_rv))
                    )
    if args.outallreads != None:
        with open(args.outallreads, "w") as f:
            f.write(
                "%s\t%s\t%s\t%s\t%s\n"
                % ("contig", "start", "start_mate", "is_fw", "mate_is_rv")
            )
            for c in allcontig_reads_d.keys():
                for read_info in allcontig_reads_d[c]:
                    s, s_m, read_is_fw, mate_is_rv = read_info
                    f.write(
                        "%s\t%i\t%i\t%s\t%s\n"
                        % (c, s, s_m, str(read_is_fw), str(mate_is_rv))
                    )



#!/usr/bin/env python3

import sys

import rich_click as click
from loguru import logger

from PlasMAAG.click_file_types import OneOrMoreSnakemakeArguments, WssFile
from PlasMAAG.command_line_runners import CliRunner, SnakemakeRunner
from PlasMAAG.environment import EnvironmentManager
from PlasMAAG.richclick_options import *


@click.command()
@click.option(
    "-r",
    "--reads",
    help="""\bWhite space separated file containing read pairs. 
<Notice the header names are required to be: read1 and read2>
This file could look like:
```
read1                     read2
path/to/sample_1/read1    path/to/sample_1/read2
path/to/sample_2/read1    path/to/sample_2/read2
```
Passing in this file means that the pipeline will be run from the start, meaning it will also assemble the reads.

""",
    type=WssFile(
        expected_headers=["read1", "read2"],
        none_file_columns=[],
    ),
)
@click.option(
    "-a",
    "--reads_and_assembly_dir",
    help=f"""\bWhite space separated file containing read pairs and paths to Spades output assembly directories.
<Notice the header names are required to be: read1, read2 and assembly_dir>
This file could look like:  
```
read1                  read2                  assembly_dir
path/sample_1/read1    path/sample_1/read2    path/sample_1/Spades_dir
path/sample_2/read1    path/sample_2/read2    path/sample_2/Spades_dir
```
Passing in this file means that the pipeline will not assemble the reads but run everything after the assembly step. 
        """,
    type=WssFile(
        expected_headers=[
            "read1",
            "read2",
            "assembly_dir",
        ],
        spades_column="assembly_dir",
    ),
)
@click.option(
    "-t",
    "--threads",
    help="Number of threads to run the application with",
    show_default=True,
    type=int,
)
@click.option(
    "-o",
    "--output",
    help="Output directory for the files produced by the pipeline",
    type=click.Path(exists=False),
)
@click.option(
    "-d",
    "--genomad_db",
    help="Path to the genomad database, if it's already installed",
    type=click.Path(exists=False),
)
@click.option(
    "-e",
    "--setup_env",
    help="Setup environment, this will be done automatically the first time the application is run",
    is_flag=True,
)
@click.option(
    "-n",
    "--dryrun",
    help="Run a snakemake dryrun for the specified files. Showing the parts of the pipeline which will be run ",
    is_flag=True,
)
@click.option(
    "-nn",
    "--cli_dryrun",
    help="Run a dryrun for the cli interface. Showing the commands which would be run from the cli interface",
    is_flag=True,
)
@click.option(
    "-s",
    "--snakemake_arguments",
    type=OneOrMoreSnakemakeArguments(),
    help="String of white space seperated snakemake arguments. eg. PlasMAAG <options> --snakemake_arguments '-n -p'",
)
@click.option(
    "-v",
    "--vamb_arguments",
    help="Arguments to pass to vamb",
)
def main(
    setup_env,
    reads,
    threads,
    dryrun,
    reads_and_assembly_dir,
    output,
    cli_dryrun,
    snakemake_arguments,
    genomad_db,
    vamb_arguments
):
    """
    \bPlasMAAG is a tool to recover plasmids and organisms from metagenomic samples.
    The first time running the program it will try to install the genomad database (~3.1 G) and required tools.
    For running the pipeline either the --reads or the --reads_and_assembly_dir arguments are required.
    Additionally, the --output argument is required (which defines the output directory) and the --threads argument.
    For Quick Start please see the README: https://github.com/RasmussenLab/PlasMAAG
    """

    if cli_dryrun:
        # Default is False
        CliRunner.dry_run_command = True

    environment_manager = EnvironmentManager()

    if genomad_db is not None:
        environment_manager._genomad_db_exist = True

    # Set up the environment
    if setup_env and not dryrun:
        environment_manager.setup_environment()
        sys.exit()

    if output is None:
        raise click.BadParameter(
            "--output is required",
        )

    if threads is None:
        raise click.BadParameter(
            "--threads is a required argument",
        )

    if reads_and_assembly_dir is not None and reads is not None:
        raise click.BadParameter(
            "Both --reads_and_assembly and --reads are used, only use one of them",
        )

    if reads_and_assembly_dir is None and reads is None:
        raise click.BadParameter(
            "Neither --reads_and_assembly and --reads are used, please define one of them",
        )

    # Check if the environment is setup correctly, if not set it up
    if not environment_manager.check_if_everything_is_setup() and not dryrun:
        environment_manager.setup_environment()

    snakemake_runner = SnakemakeRunner(snakefile="snakefile.smk")
    snakemake_runner.add_arguments(["-c", str(threads)])
    snakemake_runner.add_arguments(["-p"]) # Print out verbose information about snakemake run

    if vamb_arguments is not None:
        snakemake_runner.add_to_config(f"vamb_arguments={vamb_arguments}")

    if genomad_db is not None:
        logger.info(f"Setting genomad database path to {genomad_db}")
        snakemake_runner.add_to_config(f"genomad_database={genomad_db}")

    if snakemake_arguments is not None:
        logger.info(f"Expanding snakemake arguments with: {snakemake_arguments}")
        snakemake_runner.add_arguments(snakemake_arguments)

    # Set output directory
    snakemake_runner.output_directory = output

    # Run the pipeline from the reads, meaning the pipeline will assemble the reads beforehand
    if reads is not None:
        snakemake_runner.add_to_config(f"read_file={reads}")
        snakemake_runner.to_print_while_running_snakemake = (
            f"Running snakemake with {threads} thread(s), from paired reads"
        )

    # Run the pipeline from the reads and the assembly graphs
    if reads_and_assembly_dir is not None:
        snakemake_runner.add_to_config(f"read_assembly_dir={reads_and_assembly_dir}")
        snakemake_runner.to_print_while_running_snakemake = f"Running snakemake with {threads} thread(s), from paired reads and assembly graph"

    if dryrun:
        snakemake_runner.add_arguments(["-n"])

    snakemake_runner.run()


if __name__ == "__main__":
    main()



import os
import subprocess


def get_git_commit(script_path):
    try:
        # Get the directory of the script
        script_dir = os.path.dirname(os.path.abspath(script_path))

        # Run the Git command in the script's directory
        commit = (
            subprocess.check_output(
                ["git", "rev-parse", "--short", "HEAD"],
                cwd=script_dir,  # Specify the working directory for the Git command
            )
            .decode("utf-8")
            .strip()
        )
        return commit
    except subprocess.CalledProcessError:
        return "No Git commit found"
    except FileNotFoundError:
        return "Git not found"



import argparse
import os
import pickle

import networkx as nx
import numpy as np
from git_commit import get_git_commit


def graph_from_blastout(
    blastfile,
    min_id,
    min_al,
    restrictive_alignments,
    separator,
):
    """
    Given a blastfile that contains blast output between contigs,
    create a nx graph, where contigs are nodes and edges exist if blast alignment above min_id identity
    over at least min_al bps, [optional] allowing only matches that are restrictive, i.e. do not allow matches where a contig
    spans another contig completely.

    """
    if not restrictive_alignments:
        print("Allowing non restrictive alignments")

    def valid_alignment(c_q_len, c_s_len, q_s, q_e, s_s, s_e):
        """
        Check if alignment is consistent with both contigs belonging to the same original genome
        """

        if q_s > s_s and s_e < (c_q_len - q_s) or s_s > q_s and q_e < (c_s_len - s_s):
            return False

        else:
            return True

    def contig_len(contigname):
        return int(contigname.split("_length_")[1].split("_")[0])

    def merge_ranges(ranges_list):
        # Step 2: Merge overlapping and adjacent ranges
        merged_ranges = []
        current_start, current_end = ranges_list[0]

        for start, end in ranges_list[1:]:
            if start <= current_end:  # Overlapping or adjacent ranges
                current_end = max(current_end, end)
            else:  # No overlap
                merged_ranges.append((current_start, current_end))
                current_start, current_end = start, end

        # Don't forget to add the last range
        merged_ranges.append((current_start, current_end))

        # Step 3: Calculate the total length of the merged ranges
        total_length = sum(end - start for start, end in merged_ranges)

        return merged_ranges, total_length

    cc_blast_g = nx.Graph()
    for blast_row in np.loadtxt(blastfile, dtype=object, delimiter="\t"):
        ci, cj = blast_row[0], blast_row[1]

        # WARNING LASSE I dont get this assert if they are the same crash?
        # But arent they aligned all against all?
        # Pehaps skip does which are the same ?
        # assert ci != cj:

        if (
            ci.split(separator)[0] == cj.split(separator)[0]
        ):  ## continue if matches from the same sample
            continue

        alig_len = int(blast_row[3])
        identity = float(blast_row[2])

        if alig_len < min_al:
            continue

        if alig_len >= min_al and identity >= min_id:
            ci_s, ci_e, cj_s, cj_e = [int(p) for p in blast_row[6:10]]
            valid_al = valid_alignment(
                contig_len(ci), contig_len(cj), ci_s, ci_e, cj_s, cj_e
            )

            if (ci, cj) not in cc_blast_g.edges():
                al_ranges_q = [(min(ci_s, ci_e), max(ci_s, ci_e))]
                al_ranges_s = [(min(cj_s, cj_e), max(cj_s, cj_e))]
                merged_ranges_q, al_len_q = (
                    al_ranges_q,
                    al_ranges_q[0][1] - al_ranges_q[0][0],
                )
                merged_ranges_s, al_len_s = (
                    al_ranges_s,
                    al_ranges_s[0][1] - al_ranges_s[0][0],
                )

                alig_len = min(al_len_q, al_len_s)

                weight = (
                    identity
                    * (
                        min(alig_len, contig_len(ci), contig_len(cj))
                        / min(contig_len(ci), contig_len(cj))
                    )
                    * (1 / 100)
                )
                cc_blast_g.add_edge(
                    ci,
                    cj,
                    al_len=alig_len,
                    identity=identity,
                    restrictive=valid_al,
                    weight=weight,
                    q=ci,
                    s=cj,
                )

                cc_blast_g[ci][cj]["al_ranges_%s" % ci] = merged_ranges_q
                cc_blast_g[ci][cj]["al_ranges_%s" % cj] = merged_ranges_s

            else:
                if cc_blast_g[ci][cj]["q"] == cj:
                    ci, cj = cj, ci
                    ci_s, ci_e, cj_s, cj_e = cj_s, cj_e, ci_s, ci_e

                al_ranges_q = cc_blast_g[ci][cj]["al_ranges_%s" % ci] + [
                    (min(ci_s, ci_e), max(ci_s, ci_e))
                ]
                al_ranges_s = cc_blast_g[ci][cj]["al_ranges_%s" % cj] + [
                    (min(cj_s, cj_e), max(cj_s, cj_e))
                ]
                al_ranges_q.sort()
                al_ranges_s.sort()

                merged_ranges_q, al_len_q = merge_ranges(al_ranges_q)
                merged_ranges_s, al_len_s = merge_ranges(al_ranges_s)
                alig_len = min(al_len_q, al_len_s)

                extended_alig_len = alig_len - cc_blast_g[ci][cj]["al_len"]

                cc_blast_g[ci][cj]["identity"] = (
                    cc_blast_g[ci][cj]["identity"] * cc_blast_g[ci][cj]["al_len"]
                    + identity * extended_alig_len
                ) / (alig_len)
                cc_blast_g[ci][cj]["al_len"] = alig_len
                weight = (
                    cc_blast_g[ci][cj]["identity"]
                    * (
                        min(
                            cc_blast_g[ci][cj]["al_len"],
                            contig_len(ci),
                            contig_len(cj),
                        )
                        / min(contig_len(ci), contig_len(cj))
                    )
                    * (1 / 100)
                )
                cc_blast_g[ci][cj]["weight"] = weight
                if not cc_blast_g[ci][cj]["restrictive"]:
                    cc_blast_g[ci][cj]["restrictive"] = valid_al

    if restrictive_alignments:
        edges_to_remove = [
            (u, v) for u, v in cc_blast_g.edges() if not cc_blast_g[u][v]["restrictive"]
        ]
        cc_blast_g.remove_edges_from(edges_to_remove)

    return cc_blast_g


if __name__ == "__main__":
    # Initialize the parser
    parser = argparse.ArgumentParser(description="Generate nx graphs from gfa files")

    # Add optional arguments with `--flags`
    parser.add_argument(
        "--blastout",
        required=True,
        type=str,
        help="Path to the blsat output file.",
    )
    parser.add_argument(
        "--out",
        dest="outfile",
        required=True,
        type=str,
        help="File where the networkx graph file will be stored.",
    )

    parser.add_argument(
        "--minid",
        type=float,
        help="Min identity allowed.",
    )

    parser.add_argument(
        "--min_al_len",
        type=int,
        help="Min alignment length allowed.",
        default=500,
        required=False,
    )

    parser.add_argument(
        "--non_restrictive",
        action="store_true",
        help="if set, alignment graph will also contain edges from non restrictive alignments.",
    )
    parser.add_argument(
        "--separator",
        type=str,
        default="C",
        help="Separator that indicates the end of the sample name and the beggining of the contig name [C].",
    )

    # Parse the arguments
    args = parser.parse_args()
    print(args)

    ## Print git commit so we can debug
    commit_hash = get_git_commit(os.path.abspath(__file__))
    print(f"Git commit hash: {commit_hash}")

    with open(args.outfile, "wb") as pkl:
        pickle.dump(
            graph_from_blastout(
                args.blastout,
                min_id=args.minid,
                min_al=args.min_al_len,
                restrictive_alignments=True if not args.non_restrictive else False,
                separator=args.separator,
            ),
            pkl,
        )



import argparse
import os

import numpy as np
from git_commit import get_git_commit

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Run n2v")

    # Add optional arguments with `--flags`
    parser.add_argument(
        "--cls_plamb",
        required=True,
        type=str,
        help="Path to the plamb clusters",
    )

    parser.add_argument(
        "--cls_circular",
        required=True,
        type=str,
        help="Path to the circular clusters",
    )

    parser.add_argument(
        "--outcls",
        required=True,
        type=str,
        help="Path were circular contigs will be dumped",
    )

    ## Print git commit so we can debug
    commit_hash = get_git_commit(os.path.abspath(__file__))
    print("Git commit hash: " + commit_hash)

    # Parse the arguments
    args = parser.parse_args()
    print(args)

    ## merge clusters
    circular_contigs_arr = np.loadtxt(args.cls_circular, dtype=object)
    if circular_contigs_arr.shape[0] == 1:
        circular_contigs = set()
    else:
        cls_plamb_arr = np.loadtxt(args.cls_plamb, skiprows=1, dtype=object)
        circular_contigs = set(circular_contigs_arr[1:, 1])

    with open(args.outcls, "w") as f:
        f.write("%s\t%s\n" % ("clustername", "contigname"))
        for c in circular_contigs:
            f.write("%s_circ\t%s\n" % (c, c))
        for cl, c in cls_plamb_arr:
            if c in circular_contigs:
                continue
            f.write("%s\t%s\n" % (cl, c))



import argparse
import os
import pickle
import re
import time
from collections import Counter, defaultdict

import networkx as nx
import numpy as np
from git_commit import get_git_commit


def load_gfa(contigs_file: str, gfa_file: str):
    """
    Given the contigs.paths file and the gfa assembly graph file, generate a networkx graph object:
        - graph_objec: undirected graph, each contig is a node, create edge if the starting of ending segment of the contig is
        present in any other contig
    """
    paths = {}  # dictionary {contig_id: [first_segment, last_segment]}
    segment_contigs = {}  # dictionary  {segment:{contig_n,contig_n}} considering any segment part of the path
    node_count = 0
    # Get contig paths from contigs.paths
    with open(contigs_file) as file:
        name = file.readline()
        path = file.readline()
        while name != "" and path != "":
            while ";" in path:
                path = path[:-2] + "," + file.readline()

            start = "NODE_"
            end = "_length_"
            # Extract the integer between NODE_ and _length_
            contig_num = str(
                int(re.search("%s(.*)%s" % (start, end), name).group(1)) - 1
            )

            # Those are all the segments that connect all sequences composing the contig, the first and last one will be
            # used to define the beggining and end of the contig
            segments = path.rstrip().split(",")

            if contig_num not in paths:
                node_count += 1

                paths[contig_num] = [
                    segments[0],
                    segments[-1],
                ]  # first and last segment
            # add the the contig name to the segment key, so we can know into which contig a segment appears
            for segment in segments:
                if segment not in segment_contigs:
                    segment_contigs[segment] = set([contig_num])
                else:
                    segment_contigs[segment].add(contig_num)

            name = file.readline()
            path = file.readline()

    links = []  # ["46463 - 34345 +","652 + 8238 -",...]
    links_map = defaultdict(
        set
    )  # dictionary {segment:{segment_i,segment_j}}, indicates mapping between segments
    links_map_e2s = defaultdict(
        set
    )  # same than links_map but order consistent, end 2 start
    links_map_s2e = defaultdict(
        set
    )  # same than links_map but order consistent, start 2 end

    # Get gfa file from
    with open(gfa_file) as file:
        line = file.readline()
        while line != "":
            # Identify lines with link information
            if "L" in line:
                strings = line.split("\t")
                f1, f2 = strings[1] + strings[2], strings[3] + strings[4]
                links_map[f1].add(f2)
                links_map[f2].add(f1)
                links.append(strings[1] + strings[2] + " " + strings[3] + strings[4])

                links_map_e2s[f1].add(f2)
                links_map_s2e[f2].add(f1)

            line = file.readline()
    # Create graph
    graph_object = nx.Graph()

    # Add vertices
    for i in range(node_count):
        graph_object.add_node(i)  # change it later so it matches

    for i in range(len(paths)):
        # for each contig, retrieve the first and last segment, which define the beginning and end of the contig
        segments = paths[str(i)]

        start = segments[0]
        start_rev = ""
        # if the start segment is not complement (+), the start of the reverse will be the same but with (-)
        if start.endswith("+"):
            start_rev = start[:-1] + "-"
        else:
            start_rev = start[:-1] + "+"

        end = segments[1]
        end_rev = ""
        if end.endswith("+"):
            end_rev = end[:-1] + "-"
        else:
            end_rev = end[:-1] + "+"

        new_links = []  # list of segments that link to either my start or end segments, both in + and -

        if start in links_map:
            new_links.extend(
                list(links_map[start])
            )  # what maps to my starting segment ?
        # if the reverse complement of the starting segment is present in links_map then add it to the new_links list
        if start_rev in links_map:
            new_links.extend(
                list(links_map[start_rev])
            )  # what maps to my starting rev compl segment ?
        if end in links_map:
            new_links.extend(list(links_map[end]))  # what maps to my ending segment ?
        if end_rev in links_map:
            new_links.extend(
                list(links_map[end_rev])
            )  # what maps to my ending rev compl segment ?

        # for each segment in new_links
        for new_link in (
            new_links
        ):  # for all segment s_i that links to either my start or ending segment
            # if the segment exists as a key in segment_contigs, which indicates that is eihter part of an ending or starting segment
            if new_link in segment_contigs:  # if this segment s_i maps to a contig
                # for each contig that contains this segment (as starting or ending segment)
                for contig in segment_contigs[new_link]:
                    # if the contig that contains this segment is not itself
                    if i != int(contig):  # if this contig is not the same than my own
                        # add an edge between the contigs
                        if not graph_object.has_edge(int(contig), i):
                            graph_object.add_edge(i, int(contig), nlinks=0)

                        graph_object[int(contig)][i]["nlinks"] += 1

    return graph_object


if __name__ == "__main__":
    # Initialize the parser
    parser = argparse.ArgumentParser(description="Generate nx graphs from gfa files")

    # Add optional arguments with `--flags`
    parser.add_argument(
        "--gfa",
        dest="gfa_file",
        required=True,
        type=str,
        help="Path to the .gfa file.",
    )
    parser.add_argument(
        "--paths",
        dest="contig_path",
        required=True,
        type=str,
        help="Path to the contig.paths file.",
    )
    parser.add_argument(
        "--out",
        dest="outfile",
        required=True,
        type=str,
        help="File for the networkx graph file will be stored.",
    )
    parser.add_argument(
        "-s",
        dest="sample",
        required=True,
        type=str,
        help="Sample name that will be prefixed to contig with separator",
    )
    parser.add_argument(
        "-m",
        dest="min_contig_size",
        type=int,
        default=2000,
        help="Min contig size, connected components without more than 1 contig larger than min_contig size will be removed from the graph.",
    )
    parser.add_argument(
        "--separator",
        type=str,
        default="C",
        help="Separator that indicates the end of the sample name and the beggining of the contig name [C].",
    )

    # Parse the arguments
    args = parser.parse_args()

    ## Print git commit so we can debug
    commit_hash = get_git_commit(os.path.abspath(__file__))
    print(f"Git commit hash: {commit_hash}")

    print(args)
    t0 = time.time()
    nx_graph = load_gfa(contigs_file=args.contig_path, gfa_file=args.gfa_file)

    # Rename nodes since now they are only named as integers
    t1 = time.time()
    print("Networkx file generated in %.2f seconds." % (t1 - t0))

    ## Weight the edges based with normalized_linkage
    # 1. Extract contig lengths from contig names
    # 2. Iterate over edges and defined edge weight
    # 3. Define a new weight scaled between 0 and 1

    # 1. Extract contig lengths from contig.paths  ########## and 2. Remove graph components if have less than 2 contigs above min_contig_size
    t2 = time.time()
    contig_len_d = dict()
    binning_contigs = set()
    contignames = []
    for line in np.loadtxt(args.contig_path, dtype=object, delimiter="\t"):
        if not line.startswith("NODE_"):
            continue
        c_id = line
        c_name = args.sample + args.separator + c_id
        c_length = int(line.split("length_")[1].split("_cov")[0])
        contig_len_d[c_name] = c_length
        if c_length >= args.min_contig_size:
            binning_contigs.add(c_name)

        if c_id.startswith("NODE_") and not c_id.endswith("'"):
            contignames.append(c_id)

    print(
        "%i/%i contigs will be considered for binning."
        % (len(binning_contigs), len(contig_len_d.keys()))
    )

    cid_originalcontignames_d = {
        i: contignames[i] for i in range(len(nx_graph.nodes()))
    }

    nx_graph_renamed = nx.relabel_nodes(
        nx_graph,
        {
            node: args.sample + args.separator + cid_originalcontignames_d[node]
            for node in nx_graph.nodes()
        },
    )

    print(args.sample, list(nx_graph_renamed.nodes())[:10])

    t3 = time.time()

    print("Sample added to node names in %.2f seconds." % (t3 - t2))

    # 2. Iterate over edges and defined edge weight

    for u, v in nx_graph_renamed.edges():
        nlinks_uv = nx_graph_renamed[u][v]["nlinks"]
        # nlinks_u = np.sum([u_d["nlinks"] for u_d in nx_graph_renamed[u].values()])
        # nlinks_v = np.sum([v_d["nlinks"] for v_d in nx_graph_renamed[v].values()])
        nx_graph_renamed[u][v]["weight_unscaled"] = nlinks_uv / np.min(
            [contig_len_d[u], contig_len_d[v]]
        )

    # 3. Define a new weight scaled between 0 and 1
    t4 = time.time()
    print("Edges weighted in %.2f seconds." % (t4 - t3))
    max_edge_weight = np.percentile(
        np.array(
            [
                nx_graph_renamed[u][v]["weight_unscaled"]
                for u, v in nx_graph_renamed.edges()
            ]
        ),
        99,
    )
    for u, v in nx_graph_renamed.edges():
        nx_graph_renamed[u][v]["weight"] = min(
            nx_graph_renamed[u][v]["weight_unscaled"] / max_edge_weight, 1.0
        )

    t5 = time.time()
    print("Edge weights scaled in %.2f seconds." % (t5 - t4))

    ## Save graph to pickle file
    with open(args.outfile, "wb") as pkl:
        pickle.dump(nx_graph_renamed, pkl)

    t6 = time.time()
    print("Graph saved in %.2f seconds." % (t6 - t5))

    print("\nNetworkx file created from gfa file in %.2f seconds." % (t6 - t0))



import os
import pickle
import time

import numpy as np
import networkx as nx
import argparse
from git_commit import get_git_commit


def merge_graphs(graphs_list):

    merged_graph = nx.Graph()

    for graph in graphs_list:
        merged_graph.add_edges_from(graph.edges(data=True))

    return merged_graph


if __name__ == "__main__":
    # Initialize the parser
    parser = argparse.ArgumentParser(
        description="Merge networkx graph files from assembly graph samples and from alignment graph."
    )

    # Add optional arguments with `--flags`
    parser.add_argument(
        "--graphs_assembly",
        required=True,
        nargs="+",
        help="Path to the graphs generated from the assembly graphs.",
    )
    parser.add_argument(
        "--graph_alignment",
        required=True,
        type=str,
        help="Path to the graph generated from the alignment of contigs.",
    )

    parser.add_argument(
        "--out",
        dest="outfile",
        required=True,
        type=str,
        help="File where the merged networkx graph file will be stored.",
    )
    # Parse the arguments
    args = parser.parse_args()

    ## Print git commit so we can debug    
    commit_hash = get_git_commit(os.path.abspath(__file__))
    print(f"Git commit hash: {commit_hash}")

    print(args)
    t0 = time.time()

    graphs_to_merge = []
    for assembly_graph_f in args.graphs_assembly:
        with open(assembly_graph_f, "rb") as picklefile:
            graphs_to_merge.append(pickle.load(picklefile))

    with open(args.graph_alignment, "rb") as picklefile:
        graphs_to_merge.append(pickle.load(picklefile))

    t1 = time.time()
    print("Graphs loaded in %.2f seconds" % (t1 - t0))
    merged_graph = merge_graphs(graphs_to_merge)

    with open(args.outfile, "wb") as picklefile:
        pickle.dump(merged_graph, picklefile)

    print(
        "Merged graph has %i(%i) nodes(edges)"
        % (len(merged_graph.nodes()), len(merged_graph.edges()))
    )
    t2 = time.time()
    print("Graphs merged in %.2f seconds" % (t2 - t1))



import os
import shutil
import subprocess
from pathlib import Path
from typing import List

import rich_click as click
from loguru import logger


class CliRunner:
    """
    Class for building and executing command-line commands.
    """

    dry_run_command = False

    def __init__(self) -> None:
        self._argument_holder = []
        self._command_has_been_added = False

    def add_command_to_run(self, command_to_run):
        if self._command_has_been_added:
            raise Exception(
                f"A command has allready been added: {self._argument_holder[0]}"
            )
        self._argument_holder = [command_to_run] + self._argument_holder
        self._command_has_been_added = True

    def add_arguments(self, arguments: List):
        self._argument_holder += arguments

    def prettyprint_args(self):
        [print(x, end=" ") for x in self._argument_holder]
        print()

    def run(self):
        if self.dry_run_command:
            logger.info("Would run:")
            logger.info(self._argument_holder)
        else:
            print("Running:")
            self.prettyprint_args()
            subprocess.run(self._argument_holder)
            print("Ran:")
            self.prettyprint_args()


class SnakemakeRunner(CliRunner):
    """
    Class to run snakemake workflows from python.
    """

    output_directory = os.getcwd()
    _snakemake_path = shutil.which("snakemake")
    _src_dir = Path(__file__).parent

    def __init__(self, snakefile: str = "snakefile.smk"):
        super().__init__()
        self._config_options = []
        self.to_print_while_running_snakemake = None
        self.add_command_to_run(self._snakemake_path)
        self.snakefile_path = Path(Path(self._src_dir) / snakefile)
        self.add_arguments(["--snakefile", self.snakefile_path])
        self.add_arguments(["--rerun-triggers", "mtime"])
        self.add_arguments(["--nolock"])
        self.validate_paths()

    def validate_paths(self):
        if not self.snakefile_path.exists():
            raise click.UsageError(
                f"Could not find snakefile, tried: {self.snakefile_path}"
            )

        if self._snakemake_path is None:
            raise click.UsageError(
                """Could not find snakemake, is it installed?
See following installation guide: https://snakemake.readthedocs.io/en/stable/getting_started/installation.html"""
            )

        if shutil.which("mamba") is None:
            logger.warning(
                "Could not find mamba installation, is the correct environment activated?"
            )
            logger.warning(
                "Defaulting to use conda to build environments for snakemake, this will be slower"
            )
            self.add_arguments(["--conda-frontend", "conda"])

    def add_to_config(self, to_add):
        self._config_options += [to_add]

    def run(self):
        # use conda: always
        self.add_arguments(["--use-conda"])
        # Always rerun incomplete
        self.add_arguments(["--rerun-incomplete"])
        # Always keep incomplete
        self.add_arguments(["--keep-incomplete"])

        self.add_to_config(f"output_directory={self.output_directory}")
        # Add config options -- these needs to be passed together to snakemake
        # eg. --config <arg1> <arg2> Therefore we collect them in the self._config_options variable and
        # append them at the end
        self.add_arguments((["--config"] + self._config_options))
        # Possibility of adding info to snakemake for when it runs
        if self.to_print_while_running_snakemake is not None:
            logger.info(self.to_print_while_running_snakemake)
        # Run snakemake
        super().run()



import rich_click as click

# Setting rich options
click.rich_click.OPTION_GROUPS = {
    "PlasMAAG": [
        {
            "name": "Defining input files: One of these options are required",
            "options": ["--reads", "--reads_and_assembly_dir"],
        },
        {
            "name": "Additional required Arguments",
            "options": [
                "--output",
                "--threads",
            ],
            "table_styles": {
                "row_styles": ["yellow", "default", "default", "default"],
            },
        },
        {
            "name": "Other Options",
            "options": [
                "--dryrun",
                "--cli_dryrun",
                "--snakemake_arguments",
                "--setup_env",
                "--genomad_db",
                "--vamb_arguments",
                "--help",
            ],
            "table_styles": {
                "row_styles": ["yellow", "default", "default", "default"],
            },
        },
    ],
}

click.rich_click.USE_RICH_MARKUP = True

# Make both -h and --help available instead of just --help
CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])



from typing import Optional, IO, Union
from pathlib import Path
import vamb
import vamb.vambtools as _vambtools
#from vamb.cluster import _normalize, _calc_distances
import time
from torch.utils.data.dataset import TensorDataset as _TensorDataset
from torch.utils.data import DataLoader as _DataLoader
from torch.nn.functional import softmax as _softmax
from torch.optim import Adam as _Adam
from torch import Tensor
from torch import nn as _nn
from math import log as _log
import torch.nn.functional as F
from torch.optim.lr_scheduler import LambdaLR
from torch.utils.data import Dataset
from loguru import logger


import sys


__doc__ = """Encode a depths matrix and a tnf matrix to latent representation.

Creates a variational autoencoder in PyTorch and tries to represent the depths
and tnf in the latent space under gaussian noise.

Usage:
>>> vae = VAE(nsamples=6)
>>> dataloader, mask = make_dataloader(depths, tnf, lengths)
>>> vae.trainmodel(dataloader)
>>> latent = vae.encode(dataloader) # Encode to latent representation
>>> latent.shape
(183882, 32)
"""

__cmd_doc__ = """Encode depths and TNF using a VAE to latent representation"""

import numpy as _np
import torch as _torch


def set_batchsize(
    data_loader: _DataLoader, batch_size: int, encode=False
) -> _DataLoader:
    """Effectively copy the data loader, but with a different batch size.

    The `encode` option is used to copy the dataloader to use before encoding.
    This will not drop the last minibatch whose size may not be the requested
    batch size, and will also not shuffle the data.
    """
    return _DataLoader(
        dataset=data_loader.dataset,
        batch_size=batch_size,
        shuffle=not encode,
        drop_last=not encode,
        num_workers=1 if encode else data_loader.num_workers,
        pin_memory=data_loader.pin_memory,
    )


def make_dataloader_n2v(
    rpkm: _np.ndarray,
    tnf: _np.ndarray,
    neighs_mask: _np.ndarray,
    neighs: _np.ndarray,
    lengths: _np.ndarray,
    batchsize: int = 256,
    destroy: bool = False,
    cuda: bool = False,
) -> tuple[_DataLoader[tuple[Tensor, Tensor, Tensor, Tensor,Tensor,Tensor]], _np.ndarray]:
    """Create a DataLoader and a contig mask from RPKM and TNF.

    The dataloader is an object feeding minibatches of contigs to the VAE.
    The data are normalized versions of the input datasets, with zero-contigs,
    i.e. contigs where a row in either TNF or RPKM are all zeros, removed.
    The mask is a boolean mask designating which contigs have been kept.

    Inputs:
        rpkm: RPKM matrix (N_contigs x N_samples)
        tnf: TNF matrix (N_contigs x N_TNF)
        neighs_mask: boolean vector of len == N_contigs, where True if contig has neighs withing distance on embedding space and if contig had neighbours in the assembly-alignment graph.
        neighs: list of lists of len == N_contigs where each element indicates the neighbouring contig indices
        lenghts: contig lenghts
        batchsize: Starting size of minibatches for dataloader
        destroy: Mutate rpkm and tnf array in-place instead of making a copy.
        cuda: Pagelock memory of dataloader (use when using GPU acceleration)

    Outputs:
        DataLoader: An object feeding data to the VAE
        mask: A boolean mask of which contigs are kept
    """

    if (
        not isinstance(rpkm, _np.ndarray)
        or not isinstance(tnf, _np.ndarray)
        or not isinstance(neighs_mask, _np.ndarray)
        or not isinstance(neighs, _np.ndarray)
    ):
        raise ValueError(
            "TNF, RPKM, neighs, and neighs mask must be Numpy arrays"
        )

    if batchsize < 1:
        raise ValueError(f"Batch size must be minimum 1, not {batchsize}")

    if (
        len(rpkm) != len(tnf)
        or len(tnf) != len(lengths)
        or len(neighs_mask) != len(lengths)
        or len(neighs) != len(lengths)
    ):
        raise ValueError(
            "Lengths of TNF, RPKM, neighs, neighs mask, and lengths arrays must be the same"
        )

    if not (
        rpkm.dtype
        == tnf.dtype
        == _np.float32
    ):
        raise ValueError(
            "TNF, and RPKM  must be Numpy arrays of dtype float32"
        )

    ### Copy arrays and mask them ###
    # Copy if not destroy - this way we can have all following operations in-place
    # for simplicity
    if not destroy:
        rpkm = rpkm.copy()
        tnf = tnf.copy()
        neighs_mask = neighs_mask.copy()
        neighs = neighs.copy()

    # Normalize samples to have same depth
    sample_depths_sum = rpkm.sum(axis=0)
    if _np.any(sample_depths_sum == 0):
        raise ValueError(
            "One or more samples have zero depth in all sequences, so cannot be depth normalized"
        )
    rpkm *= 1_000_000 / sample_depths_sum

    zero_tnf = tnf.sum(axis=1) == 0
    smallest_index = _np.argmax(zero_tnf)
    if zero_tnf[smallest_index]:
        raise ValueError(
            f"TNF row at index {smallest_index} is all zeros. "
            + "This implies that the sequence contained no 4-mers of A, C, G, T or U, "
            + "making this sequence uninformative. This is probably a mistake. "
            + "Verify that the sequence contains usable information (e.g. is not all N's)"
        )

    total_abundance = rpkm.sum(axis=1)


    ## for the contigs that are present only in one sample, define the object of abundances 
    # across samples (n_contigs x n_samples) z-score normalizing per sample only for contigs that satisfy (BOTH):
    # a) abundance only in one sample
    # b) abundance in the specific sample

    rpkm_unz=_np.zeros_like(rpkm)    

    
    rpkm_unz_mask = _np.zeros(rpkm.shape[0],dtype=bool)
    if rpkm.shape[1] <= 10:
        singlesample_contig_idxs = _np.where(_np.count_nonzero(rpkm, axis=1) == 1)[0]    
        rpkm_unz_mask[singlesample_contig_idxs]=True


        for i in range(rpkm.shape[1]):
            mask_nonzeroab_sample_i = rpkm[:,i] > 0.0 
            mask_single_ab_and_nonzeroab_si = rpkm_unz_mask & mask_nonzeroab_sample_i    
            if _np.sum((mask_single_ab_and_nonzeroab_si)) == 0:
                continue
            
            
            mean_ab_s_i = _np.mean(rpkm[mask_single_ab_and_nonzeroab_si,i])
            std_ab_s_i = _np.std(rpkm[mask_single_ab_and_nonzeroab_si,i])
            rpkm_unz[mask_single_ab_and_nonzeroab_si,i] = (rpkm[mask_single_ab_and_nonzeroab_si,i]-mean_ab_s_i)/std_ab_s_i
            
            #log_vals = _np.log(rpkm[mask_single_ab_and_nonzeroab_si,i])    
            #rpkm_unz[mask_single_ab_and_nonzeroab_si,i] = (log_vals-_np.mean(log_vals))/_np.std(log_vals)

        
        #rpkm_unz=_np.log(rpkm_unz.clip(min=0.001))
        #logger.info("Adding max(log(rpkm_unz),0.001) as input")
        logger.info("Adding rpkm z-score vector as input for contigs only present in one sample %i/%i" %(_np.sum(rpkm_unz_mask),len(rpkm_unz_mask)))
    

    else:
        singlesample_contig_idxs = []
            
    # Normalize rpkm to sum to 1
    n_samples = rpkm.shape[1]
    zero_total_abundance = total_abundance == 0
    rpkm[zero_total_abundance] = 1 / n_samples
    nonzero_total_abundance = total_abundance.copy()
    nonzero_total_abundance[zero_total_abundance] = 1.0
    rpkm /= nonzero_total_abundance.reshape((-1, 1))
    
    ## for the contigs that are only present in one sample, remove the sample correltaion signal ###### NOVELTY#########
    singlesample_contig_idxs = _np.where(_np.count_nonzero(rpkm, axis=1) == 1)[0]
    rpkm[singlesample_contig_idxs,:]=0
    logger.info("Removing rpkm signal for contigs present only in one sample")


    # Normalize TNF and total abundance to make SSE loss work better
    total_abundance = _np.log(total_abundance.clip(min=0.001))
    _vambtools.zscore(total_abundance, inplace=True)
    _vambtools.zscore(tnf, axis=0, inplace=True)
    total_abundance.shape = (len(total_abundance), 1)

    # Create weights
    lengths = (lengths).astype(_np.float32)
    weights = _np.log(lengths).astype(_np.float32) - 5.0
    weights[weights < 2.0] = 2.0
    weights *= len(weights) / weights.sum()
    weights.shape = (len(weights), 1)
    indices = _np.arange(tnf.shape[0])

    ### Create final tensors and dataloader ###
    depthstensor = _torch.from_numpy(rpkm)  # this is a no-copy operation
    depthstensor_unz = _torch.from_numpy(rpkm_unz)  # this is a no-copy operation
    depthstensor_unz_mask = _torch.from_numpy(rpkm_unz_mask)  # this is a no-copy operation
    tnftensor = _torch.from_numpy(tnf)
    total_abundance_tensor = _torch.from_numpy(total_abundance)
    weightstensor = _torch.from_numpy(weights)
    neighs_masktensor = _torch.from_numpy(neighs_mask)
    indicestensor = _torch.from_numpy(indices)

    n_workers = 4 if cuda else 1
    dataset = _TensorDataset(
        depthstensor,
        tnftensor,
        total_abundance_tensor,
        neighs_masktensor,
        weightstensor,
        indicestensor,
        depthstensor_unz,
        depthstensor_unz_mask,
    )

    dataloader = _DataLoader(
        dataset=dataset,
        batch_size=batchsize,
        drop_last=False,
        shuffle=True,
        num_workers=n_workers,
        pin_memory=cuda,
    )

    return dataloader, neighs


class VAE(_nn.Module):
    """Variational autoencoder, subclass of torch.nn.Module.

    Instantiate with:
        nsamples: Number of samples in abundance matrix
        ncontigs: Number of contigs present in the catalog
        neighs_object: Array of list containing the indexes of the neighbouring contigs
        nhiddens: list of n_neurons in the hidden layers [None=Auto]
        nlatent: Number of neurons in the latent layer [32]
        alpha: Approximate starting TNF/(CE+TNF) ratio in loss. [None = Auto]
        beta: Multiply KLD by the inverse of this value [200]
        gamma: Weight for the contrastive loss term [1.0]
        dropout: Probability of dropout on forward pass [0.2]
        cuda: Use CUDA (GPU accelerated training) [False]
        seed: Seed set from where to sample random numbers

    vae.trainmodel(dataloader, nepochs batchsteps, lrate, logfile, modelfile)
        Trains the model, returning None

    vae.encode(self, data_loader):
        Encodes the data in the data loader and returns the encoded matrix.

    If alpha or dropout is None and there is only one sample, they are set to
    0.99 and 0.0, respectively
    """

    def __init__(
        self,
        nsamples: int,
        ncontigs: int,
        neighs_object: _np.ndarray,
        nhiddens: Optional[list[int]] = None,
        nlatent: int = 32,
        alpha: Optional[float] = None,
        beta: float = 200.0,
        gamma: float = 0.1,
        dropout: Optional[float] = 0.2,
        cuda: bool = False,
        seed: int = 0,
        margin: float = 0.01,
    ):
        if nlatent < 1:
            raise ValueError(f"Minimum 1 latent neuron, not {nlatent}")

        if nsamples < 1:
            raise ValueError(f"nsamples must be > 0, not {nsamples}")

        # If only 1 sample, we weigh alpha and nhiddens differently
        if alpha is None:
            alpha = 0.15 if nsamples > 1 else 0.50

        if nhiddens is None:
            nhiddens = [512, 512] if nsamples > 1 else [256, 256]

        if dropout is None:
            dropout = 0.2 if nsamples > 1 else 0.0

        if any(i < 1 for i in nhiddens):
            raise ValueError(f"Minimum 1 neuron per layer, not {min(nhiddens)}")

        if beta <= 0:
            raise ValueError(f"beta must be > 0, not {beta}")

        if not (0 < alpha < 1):
            raise ValueError(f"alpha must be 0 < alpha < 1, not {alpha}")

        if not (0 <= dropout < 1):
            raise ValueError(f"dropout must be 0 <= dropout < 1, not {dropout}")

        _torch.manual_seed(seed)
        self.rng = _torch.Generator()
        self.rng.manual_seed(seed)
        super(VAE, self).__init__()

        # Initialize simple attributes
        self.usecuda = cuda
        self.nsamples = nsamples
        self.ncontigs = ncontigs

        self.ntnf = 103
        self.alpha = alpha
        self.beta = beta
        self.gamma = gamma
        self.nhiddens = nhiddens
        self.nlatent = nlatent
        self.dropout = dropout
        self.margin = margin
        


        # Initialize lists for holding hidden layers
        self.encoderlayers = _nn.ModuleList()
        self.encodernorms = _nn.ModuleList()
        self.decoderlayers = _nn.ModuleList()
        self.decodernorms = _nn.ModuleList()

        # Add all other hidden layers
        for nin, nout in zip(
            [self.nsamples + self.ntnf + 1 + self.nsamples] + self.nhiddens,
            self.nhiddens,
        ):
            self.encoderlayers.append(_nn.Linear(nin, nout))
            self.encodernorms.append(_nn.BatchNorm1d(nout))

        # Latent layers
        self.mu = _nn.Linear(self.nhiddens[-1], self.nlatent)

        # Add first decoding layer
        for nin, nout in zip([self.nlatent] + self.nhiddens[::-1], self.nhiddens[::-1]):
            self.decoderlayers.append(_nn.Linear(nin, nout))
            self.decodernorms.append(_nn.BatchNorm1d(nout))

        # Reconstruction (output) layer
        self.outputlayer = _nn.Linear(
            self.nhiddens[0], self.nsamples + self.ntnf + 1 + self.nsamples  # + self.n_embedding
        )

        # Activation functions
        self.relu = _nn.LeakyReLU()
        self.softplus = _nn.Softplus()
        self.dropoutlayer = _nn.Dropout(p=self.dropout)

        # Object where I will store the mu s , so we update after training
        
        self.mu_container = _torch.randn(
            self.ncontigs, self.nlatent, dtype=_torch.float32
        )
        
        # Container where I know the neighbours of each contig in embedding space
        self.neighs = neighs_object
        if self.usecuda:
            self.cuda()
            self.mu_container = self.mu_container.cuda()

    def _encode(self, tensor: Tensor) -> Tensor:
        tensors = list()

        # Hidden layers
        for encoderlayer, encodernorm in zip(self.encoderlayers, self.encodernorms):
            tensor = encodernorm(self.dropoutlayer(self.relu(encoderlayer(tensor))))
            tensors.append(tensor)

        # Latent layers
        mu = self.mu(tensor)

        # Note: We ought to also compute logsigma here, but we had a bug in the original
        # implementation of Vamb where logsigma was fixed to zero, so we just remove it.

        return mu

    # sample with gaussian noise
    def reparameterize(self, mu: Tensor) -> Tensor:
        epsilon = _torch.randn(mu.size(0), mu.size(1))

        if self.usecuda:
            epsilon = epsilon.cuda()

        epsilon.requires_grad = True

        latent = mu + epsilon

        return latent

    def _decode(self, tensor: Tensor) -> tuple[Tensor, Tensor]:
        tensors = list()

        for decoderlayer, decodernorm in zip(self.decoderlayers, self.decodernorms):
            tensor = decodernorm(self.dropoutlayer(self.relu(decoderlayer(tensor))))
            tensors.append(tensor)

        reconstruction = self.outputlayer(tensor)

        # Decompose reconstruction to depths and tnf signal
        depths_out = reconstruction.narrow(1, 0, self.nsamples)
        tnf_out = reconstruction.narrow(1, self.nsamples, self.ntnf)
        abundance_out = reconstruction.narrow(1, self.nsamples + self.ntnf, 1)
        abundance_long_out = reconstruction.narrow(1, self.nsamples + self.ntnf + 1, self.nsamples)
        
        # emb_out = reconstruction.narrow(1, self.ntnf + self.nsamples, self.n_embedding)
        # If multiple samples, apply softmax
        if self.nsamples > 1:
            depths_out = _softmax(depths_out, dim=1)
        
        return depths_out, tnf_out, abundance_out ,abundance_long_out

    def forward(
        self,
        depths: Tensor,
        tnf: Tensor,
        abundance: Tensor,
        abundance_long: Tensor,
    ) -> tuple[Tensor, Tensor, Tensor, Tensor]:
        tensor = _torch.cat((depths, tnf, abundance,abundance_long), 1)
        mu = self._encode(tensor)
        latent = self.reparameterize(mu)
        depths_out, tnf_out, abundance_out,abundance_long_out = self._decode(latent)

        return depths_out, tnf_out, abundance_out,abundance_long_out, mu

    def calc_loss(
        self,
        depths_in: Tensor,
        depths_out: Tensor,
        tnf_in: Tensor,
        tnf_out: Tensor,
        abundance_in: Tensor,
        abundance_out: Tensor,
        mu: Tensor,
        weights: Tensor,
        preds_idxs: Tensor,
        neighs_mask: Tensor,
        abundance_long_in: Tensor,
        abundance_long_out: Tensor,
        abundance_long_mask: Tensor,
    ) -> tuple[Tensor, Tensor, Tensor, Tensor, Tensor,Tensor]:
        
        
        # If multiple samples, use cross entropy, else use SSE for abundance
        #t0=time.time()
        ab_sse = (abundance_out - abundance_in).pow(2).sum(dim=1)
        ab_sse_weight = (1 - self.alpha) * (1 / self.nsamples)
        weighed_ab = ab_sse * ab_sse_weight
        #time_ab_sse=time.time()-t0
        
        #t0=time.time()
        ce = -((depths_out + 1e-9).log() * depths_in).sum(dim=1)
        # Avoid having the denominator be zero
        if self.nsamples == 1:
            ce_weight = 0.0
        else:
            ce_weight = ((1 - self.alpha) * (self.nsamples - 1)) / (
                self.nsamples * _log(self.nsamples)
            )
        weighed_ce = ce * ce_weight        
        #time_ce=time.time()-t0
        
        #t0=time.time()
        sse = (tnf_out - tnf_in).pow(2).sum(dim=1)
        sse_weight = self.alpha / self.ntnf
        weighed_sse = sse * sse_weight
        #time_sse=time.time()-t0
        
        #t0=time.time()
        sse_ab_long = (abundance_long_out - abundance_long_in).pow(2).sum(dim=1)*abundance_long_mask 
        ab_long_sse_weight = self.alpha / (self.nsamples)
        weighed_ab_long_sse = sse_ab_long * ab_long_sse_weight
        #time_sse_long=time.time()-t0
        
        #t0=time.time()
        kld = 0.5 * (mu.pow(2)).sum(dim=1)
        kld_weight = 1 / (self.nlatent * self.beta)
        weighed_kld = kld * kld_weight
        #time_kld=time.time()-t0
        
        #t0=time.time()
        loss_emb = self.cosinesimilarity_loss(
            mu,
            preds_idxs,
            neighs_mask,
        )
        loss_emb_cat = _torch.cat(
            (
                loss_emb.unsqueeze(0) - self.margin,
                _torch.zeros_like(loss_emb).unsqueeze(0),
            ),
            dim=0,
        )
        #time_contr=time.time()-t0

        reconstruction_loss = (
            weighed_ce
            + weighed_ab
            + weighed_sse
            + weighed_ab_long_sse
        )
        loss = (reconstruction_loss + weighed_kld) * weights + _torch.max(
            loss_emb_cat, dim=0
        )[0] * self.gamma

        loss_emb_pop = loss_emb[neighs_mask]
        
        return (
            loss.mean(),
            weighed_ab.mean(),
            weighed_ce.mean(),
            weighed_sse.mean(),
            (weighed_ab_long_sse[abundance_long_mask]).mean(),
            weighed_kld.mean(),
            loss_emb_pop.mean(), 
            #(time_ab_sse,time_ce,time_sse,time_sse_long,time_kld,time_contr)

        )

    def cosinesimilarity_loss(self, mu, idxs_preds, neighs_mask):
        results = []

        for mu_i, idx_pred, neighs_mask_i in zip(mu, idxs_preds, neighs_mask):
            if not neighs_mask_i:
                avg_cosine_distance = _torch.tensor(0.0, device=mu.device)
            else:
                mus_neighs = self.mu_container[self.neighs[idx_pred]].to(mu.device)
                cosine_distances = 2 * self.calc_cosine_distances(mus_neighs, mu_i)
                avg_cosine_distance = cosine_distances.mean()

            results.append(avg_cosine_distance)

        return _torch.stack(results)
    
    def calc_cosine_distances(self,matrix, vector):
        "Return vector of cosine distances from rows of normalized matrix to given row."
        # normalize vector 
        vector = vector/(vector.norm() * (2**0.5))
        
        dists = 0.5 - matrix.matmul(vector)
        
        return dists

    def trainepoch(
        self,
        data_loader: _DataLoader,
        epoch: int,
        optimizer,
        batchsteps: list[int],
    ) -> _DataLoader[tuple[Tensor, Tensor, Tensor, Tensor, Tensor, Tensor]]:
        self.train()

        epoch_loss = 0.0
        epoch_kldloss = 0.0
        epoch_sseloss = 0.0
        epoch_celoss = 0.0
        epoch_embloss_pop = 0.0
        epoch_absseloss = 0.0
        epoch_ablongsseloss = 0.0

        
        # epoch_kldloss_time = 0.0
        # epoch_sseloss_time = 0.0
        # epoch_celoss_time = 0.0
        # epoch_embloss_pop_time = 0.0
        # epoch_absseloss_time = 0.0
        # epoch_ablongsseloss_time = 0.0


        if epoch in batchsteps:
            data_loader = set_batchsize(data_loader, data_loader.batch_size * 2)

        for (
            depths_in,
            tnf_in,
            abundance_in,
            neighs_mask,
            weights,
            preds_idxs,
            abundance_long_in,
            abundance_long_mask,
        ) in data_loader:
            depths_in.requires_grad = True
            tnf_in.requires_grad = True
            
    
            preds_idxs = preds_idxs.long()
            neighs_mask = neighs_mask.bool()
            abundance_long_mask = abundance_long_mask.bool()
            
            if self.usecuda:
                depths_in = depths_in.cuda()
                tnf_in = tnf_in.cuda()
                abundance_in = abundance_in.cuda()
                neighs_mask = neighs_mask.cuda()
                weights = weights.cuda()
                preds_idxs = preds_idxs.cuda()
                abundance_long_in = abundance_long_in.cuda()
                abundance_long_mask = abundance_long_mask.cuda()

            optimizer.zero_grad()

            depths_out, tnf_out, abundance_out,abundance_long_out, mu = self(
                depths_in, tnf_in, abundance_in,abundance_long_in
            )
            
            self.mu_container[preds_idxs] = vamb.cluster._normalize(mu.detach())
            (
                loss,
                ab_sse,
                ce,
                sse,
                ab_long_sse,
                kld,
                loss_emb_pop,
                #times
            ) = self.calc_loss(
                depths_in,
                depths_out,
                tnf_in,
                tnf_out,
                abundance_in,
                abundance_out,
                mu,
                weights,
                preds_idxs,
                neighs_mask,
                abundance_long_in,
                abundance_long_out,
                abundance_long_mask,
            )
            
            
            
            loss.backward()
            optimizer.step()
            epoch_loss += loss.data.item()
            epoch_kldloss += kld.data.item()
            epoch_sseloss += sse.data.item()
            epoch_celoss += ce.data.item()
            epoch_embloss_pop += loss_emb_pop.item() 
            epoch_absseloss += ab_sse.data.item()
            epoch_ablongsseloss += ab_long_sse.data.item()
            
        #     time_ab_sse,time_ce,time_sse,time_sse_long,time_kld,time_contr = times
            
        #     epoch_kldloss_time += time_kld
        #     epoch_sseloss_time += time_sse
        #     epoch_celoss_time += time_ce
        #     epoch_embloss_pop_time += time_contr
        #     epoch_absseloss_time += time_ab_sse
        #     epoch_ablongsseloss_time += time_sse_long
        
        # epoch_loss_time = epoch_kldloss_time + epoch_sseloss_time + epoch_celoss_time + epoch_embloss_pop_time + epoch_absseloss_time + epoch_ablongsseloss_time
        #logger.info("monitoring runtimes")
        logger.info(
            "\tEpoch: {}\tLoss: {:.6f}\tCE: {:.6f}\tAB:{:.4e}\tABlong:{:.4e}\tSSE: {:.6f}\tembloss_pop: {:.6f}\tKLD: {:.4f}\tBatchsize: {}".format(
                epoch + 1,
                epoch_loss / len(data_loader),
                epoch_celoss / len(data_loader),
                epoch_absseloss / len(data_loader),
                epoch_ablongsseloss / len(data_loader),
                epoch_sseloss / len(data_loader),
                epoch_embloss_pop / len(data_loader),
                epoch_kldloss / len(data_loader),

                # epoch_loss_time,
                # epoch_celoss_time,
                # epoch_absseloss_time,
                # epoch_ablongsseloss_time,
                # epoch_sseloss_time,
                # epoch_embloss_pop_time,
                # epoch_kldloss_time,
                
                data_loader.batch_size,
                
                
                
            )
        )
        
        self.eval()
        return data_loader

    def encode(self, data_loader) -> _np.ndarray:
        """Encode a data loader to a latent representation with VAE

        Input: data_loader: As generated by train_vae

        Output: A (n_contigs x n_latent) Numpy array of latent repr.
        """

        self.eval()

        new_data_loader = set_batchsize(
            data_loader, data_loader.batch_size, encode=True
        )

        depths_array, _, _, _, _, _,_,_ = data_loader.dataset.tensors
        length = len(depths_array)

        # We make a Numpy array instead of a Torch array because, if we create
        # a Torch array, then convert it to Numpy, Numpy will believe it doesn't
        # own the memory block, and array resizes will not be permitted.
        latent = _np.empty((length, self.nlatent), dtype=_np.float32)

        row = 0
        with _torch.no_grad():
            for depths, tnf, ab, _, _, _,ab_long,_ in new_data_loader:
                # Move input to GPU if requested
                if self.usecuda:
                    depths = depths.cuda()
                    tnf = tnf.cuda()
                    ab = ab.cuda()
                    ab_long = ab_long.cuda()

                # Evaluate'
                _, _, _,_,mu = self(depths, tnf, ab,ab_long)
                

                if self.usecuda:
                    mu = mu.cpu()

                latent[row : row + len(mu)] = mu
                row += len(mu)

        assert row == length
        return latent

    def save(self, filehandle):
        """Saves the VAE to a path or binary opened file. Load with VAE.load

        Input: Path or binary opened filehandle
        Output: None
        """
        state = {
            "nsamples": self.nsamples,
            "ncontigs": self.ncontigs,
            "alpha": self.alpha,
            "beta": self.beta,
            "gamma": self.gamma,
            "dropout": self.dropout,
            "nhiddens": self.nhiddens,
            "nlatent": self.nlatent,
            "margin": self.margin,
            "state": self.state_dict(),
        }

        _torch.save(state, filehandle)


    @classmethod
    def load(
        cls, path: Union[IO[bytes], str], cuda: bool = False, evaluate: bool = True
    ):
        """Instantiates a VAE from a model file.
        Inputs:
            path: Path to model file as created by functions VAE.save or
                  VAE.trainmodel.
            cuda: If network should work on GPU [False]
            evaluate: Return network in evaluation mode [True]

        Output: VAE with weights and parameters matching the saved network.
        """
        
        # Forcably load to CPU even if model was saves as GPU model
        dictionary = _torch.load(path, map_location=lambda storage, loc: storage)

        nsamples = dictionary["nsamples"]
        alpha = dictionary["alpha"]
        beta = dictionary["beta"]
        dropout = dictionary["dropout"]
        nhiddens = dictionary["nhiddens"]
        nlatent = dictionary["nlatent"]
        state = dictionary["state"]

        vae = cls(nsamples, nhiddens, nlatent, alpha, beta, dropout, cuda)
        vae.load_state_dict(state)

        if cuda:
            vae.cuda()

        if evaluate:
            vae.eval()

        return vae

    def trainmodel(
        self,
        dataloader: _DataLoader[
            tuple[Tensor, Tensor, Tensor, Tensor, Tensor, Tensor, Tensor]
        ],
        nepochs: int = 500,
        lrate: float = 1e-3,
        batchsteps: Optional[list[int]] = [25, 75, 150, 300],
        modelfile: Union[None, str, Path, IO[bytes]] = None,
    ):
        """Train the autoencoder from depths array and tnf array.

        Inputs:
            dataloader: DataLoader made by make_dataloader
            nepochs: Train for this many epochs before encoding [500]
            lrate: Starting learning rate for the optimizer [0.001]
            batchsteps: None or double batchsize at these epochs [25, 75, 150, 300]
            logfile: Print status updates to this file if not None [None]
            modelfile: Save models to this file if not None [None]
            
        Output: None
        """

        if lrate < 0:
            raise ValueError(f"Learning rate must be positive, not {lrate}")

        if nepochs < 1:
            raise ValueError("Minimum 1 epoch, not {nepochs}")

        if batchsteps is None:
            batchsteps_set: set[int] = set()
        else:
            # First collect to list in order to allow all element types, then check that
            # they are integers
            batchsteps = list(batchsteps)
            if not all(isinstance(i, int) for i in batchsteps):
                raise ValueError("All elements of batchsteps must be integers")
            if max(batchsteps, default=0) >= nepochs:
                raise ValueError("Max batchsteps must not equal or exceed nepochs")
            last_batchsize = dataloader.batch_size * 2 ** len(batchsteps)
            if len(dataloader.dataset) < last_batchsize:  # type: ignore
                raise ValueError(
                    f"Last batch size of {last_batchsize} exceeds dataset length "
                    f"of {len(dataloader.dataset)}. "  # type: ignore
                    "This means you have too few contigs left after filtering to train. "
                    "It is not adviced to run Vamb with fewer than 10,000 sequences "
                    "after filtering. "
                    "Please check the Vamb log file to see where the sequences were "
                    "filtered away, and verify BAM files has sensible content."
                )
            batchsteps_set = set(batchsteps)

        # Get number of features
        # Following line is un-inferrable due to typing problems with DataLoader
        ncontigs, nsamples = dataloader.dataset.tensors[0].shape  # type: ignore
        optimizer = _Adam(self.parameters(), lr=lrate)

        logger.info(f"\tNetwork properties:")
        logger.info(f"\tCUDA: {self.usecuda}")
        logger.info(f"\tAlpha: {self.alpha}")
        logger.info(f"\tBeta: {self.beta}")
        logger.info(f"\tGamma: {self.gamma}")
        logger.info(f"\tMargin: {self.margin}")
        logger.info(f"\tDropout: {self.dropout}")
        logger.info(f"\tN hidden: {', '.join(map(str, self.nhiddens))}")
        logger.info(f"\tN latent: {self.nlatent}\n")
        logger.info(f"\tTraining properties:")
        logger.info(f"\tN epochs: {nepochs}")
        logger.info(f"\tStarting batch size: {dataloader.batch_size}")
        batchsteps_string = (
            ", ".join(map(str, sorted(batchsteps_set))) if batchsteps_set else "None"
        )
        logger.info(f"\tBatchsteps: {batchsteps_string}")
        logger.info(f"\tLearning rate: {lrate}")
        logger.info(f"\tN sequences: {ncontigs}")
        logger.info(f"\tN samples: {nsamples}\n\n")
        

        # Train
        for epoch in range(nepochs):
            dataloader = self.trainepoch(
                dataloader,
                epoch,
                optimizer,
                sorted(batchsteps_set),
            )
        # Save weights - Lord forgive me, for I have sinned when catching all exceptions
        if modelfile is not None:
            try:
                self.save(modelfile)
            except:
                pass

        return None



import argparse
import os
import pickle

# from scipy.spatial.distance import cdist
import time

import networkx as nx
import numpy as np
import torch
from git_commit import get_git_commit

import vamb


# define functions and classess
def find_neighbours_optimized(
    embeddings_bincontigs,
    contignames,
    idxs_without_embeddings,
    c_cc_d,
    radius_clustering=0.2,
):
    """ """
    radius = radius_clustering / 2

    communities_g = nx.Graph()
    embeddings_bincontigs_nz = vamb.cluster._normalize(embeddings_bincontigs)
    neighs = []

    ## For each contig embeddings_bincontigs, find the contigs that are within a distance, and link them
    for i in range(len(embeddings_bincontigs)):
        neighs.append([])
        if i in idxs_without_embeddings:
            continue

        distances = vamb.cluster._calc_distances(embeddings_bincontigs_nz, i)
        ## can I use the triangle inequality to optimize this, so it does not compute all against all for every contig?s
        mask_ = torch.ones(len(embeddings_bincontigs), dtype=torch.bool)
        mask_[i] = False
        within_radius = (distances <= radius) & mask_

        if torch.sum(within_radius) == 0:  # in such case it will not have a neighbour
            continue
        else:
            for j,lat_neigh_idx in enumerate(np.where(within_radius)[0]):
                if c_cc_d[contignames[lat_neigh_idx]] == c_cc_d[contignames[i]]:
                    communities_g.add_edge(contignames[i], contignames[lat_neigh_idx],distance=distances[within_radius][j].item())
                    neighs[i].append(lat_neigh_idx)
    
    return (np.array(neighs, dtype=object), communities_g)


def split_neighs_per_sample(neighs, contignames):
    neighs_sorted_same_sample = []
    mask_cs_with_neighs = np.zeros(len(contignames), dtype=bool)
    neighs_removed_n = 0
    neighs_kept = 0
    for i, nn_idxs in enumerate(neighs):
        if len(nn_idxs) == 0:
            neighs_sorted_same_sample.append([])
            continue

        c_i = contignames[i]
        c_i_sample = c_i.split("C")[0]
        neighs_i_same_sample_idxs = []
        for neigh_idx in nn_idxs:
            if contignames[neigh_idx].split("C")[0] == c_i_sample:
                neighs_i_same_sample_idxs.append(neigh_idx)
                neighs_kept += 1
            else:
                neighs_removed_n += 1

        neighs_sorted_same_sample.append(neighs_i_same_sample_idxs)
        if len(neighs_i_same_sample_idxs) > 0:
            mask_cs_with_neighs[i] = True
    print(
        "%i/%i neighbour connections have been removed since neighs are part of different samples."
        % (neighs_removed_n, neighs_kept + neighs_removed_n)
    )
    print(np.where(mask_cs_with_neighs)[0])
    return (
        np.array(neighs_sorted_same_sample, dtype=object),
        np.where(mask_cs_with_neighs)[0],
    )


def get_neighbourhoods(neighs_obj, contignames, min_neighs=2):
    neighbourhoods_g = nx.Graph()
    for i, neigh_idxs in enumerate(neighs_obj):
        c = contignames[i]
        for neigh_idx in neigh_idxs:
            c_neigh = contignames[neigh_idx]
            if c_neigh == c:
                continue
            neighbourhoods_g.add_edge(c_neigh, c)

    hood_cs_d = {
        i: cc
        for i, cc in enumerate(nx.connected_components(neighbourhoods_g))
        if len(cc) >= min_neighs
    }

    return (
        hood_cs_d,
        neighbourhoods_g,
    )


if __name__ == "__main__":
    # Create argument parser
    parser = argparse.ArgumentParser(
        description="From the embeddings, the list of contigs that have been embedded, and the binning contigs, generate a neighbours file, a contigs with neighbours mask, and a dictionary with neighbours and hoods."
    )
    # Add arguments
    parser.add_argument("--embs", help="Embeddings file.")
    parser.add_argument("--contigs_embs", help="Contigs embedded file.")
    parser.add_argument("--neighs_outdir", help="Neighbours outtdir.")
    parser.add_argument("-g", help="Graph files that was used as input for n2v")
    parser.add_argument(
        "--contignames",
        default=None,
        help="File containing the binning contignames, if not provided it will generate neighs with all contigs.",
    )
    parser.add_argument(
        "-r",
        default=[0.1],
        help="Radius within 2 contigs are considered neighbours.",
        nargs="+",
    )

    ## Taking the n2v embeddings from either the assembly graph or the alignment graph. Extract communities of contigs from them.
    # 1. Sort and mask so neighbours are generated according the order defined by the contignames file content
    # 2. Compute cosine distances all against all, only for considered contigs
    # 3. For each contig, define as neighbours the ones that are within a radius
    # 4. For each community of connected contigs, split community if belong to different graph componentnts

    ## Print git commit so we can debug
    commit_hash = get_git_commit(os.path.abspath(__file__))
    print(f"Git commit hash: {commit_hash}")

    ## Parse arguments
    args = parser.parse_args()
    try:
        os.mkdir(args.neighs_outdir)
    except:
        pass

    with open(os.path.join(args.neighs_outdir, "log.txt"), "w") as f:
        f.write("%s\n" % (args))

    ## Load files
    # Load graph file
    with open(args.g, "rb") as pkl_file:
        graph = pickle.load(pkl_file)
    # Define which component each contig belongs to
    ccs_graph_d = {i: cc for i, cc in enumerate(nx.connected_components(graph))}
    c_cc_d = {c: i for i, cs in ccs_graph_d.items() for c in cs}

    # Define radiuses
    radiuses = [float(r) for r in args.r]

    # Load contignames
    if args.contignames != None:
        ## Load the names of the contigs that will be used for binning BE CAREFUL SINCE HE CONTIGNAMES USUALLY USED ONLY APPLUES FOR MIN CONTIG LEN 2000
        contignames = np.loadtxt(args.contignames, dtype=object)

    else:
        contignames = np.loadtxt(args.contigs_embs, dtype=object)

    print("len of contignames %i" % (len(contignames)))
    c_idx_d = {c: i for i, c in enumerate(contignames)}

    # Load the embeddings, and the contigs that are represented in those embeddings
    contigsembs = np.loadtxt(args.contigs_embs, dtype=object)
    embeddings = np.load(args.embs)["arr_0"]
    contig_emb_d = {c: e for c, e in zip(contigsembs, embeddings)}

    ## 1. Mask the embeddings and process them so they match the binning contigs
    embeddings_bincontigs = np.zeros((len(contignames), embeddings.shape[1]))
    embeddings_mask = np.ones(len(contignames), dtype=bool)
    for i, c in enumerate(contignames):
        if c not in contig_emb_d.keys():
            embeddings_mask[i] = False
            continue
        embeddings_bincontigs[c_idx_d[c], :] = contig_emb_d[c]

    fraction_embedded_contigs = np.sum(embeddings_mask) / len(contignames)
    print("Fraction of contigs with embeddings %.3f" % (fraction_embedded_contigs))

    ## 2. Compute cosine distances all against all, only for considered AND contigs and contigs taht are embedded
    t0 = time.time()

    ## 3. For each contig, define as neighbours the ones that are within a radius
    embs_d = dict()
    for radius in radiuses:
        print("Finding neighbours within radius %.3f" % radius)
        embs_d[radius] = dict()
        embs_d[radius]["neighs"] = find_neighbours_optimized(
            embeddings_bincontigs,
            contignames,
            np.where(embeddings_mask == False)[0],
            c_cc_d,
            radius,
        )

        ## extract also neighs split by sample
        embs_d[radius]["neighs_split_by_sample"] = split_neighs_per_sample(
            embs_d[radius]["neighs"][0], contignames
        )
        ## Define community of neighbours if a path of contigs within raidus can be walked
        # embs_d[radius]["nhbds"], embs_d[radius]["nhbds_g"], hoods_split_bool = (
        embs_d[radius]["nhbds"], embs_d[radius]["nhbds_g"] = get_neighbourhoods(
            embs_d[radius]["neighs"][0], contignames
        )

        c_hood_d = {c: hood for hood, cs in embs_d[radius]["nhbds"].items() for c in cs}

        ## also save neighs where hoods are removed if they are composed of only neighs from diff samples, i.e. hoood : {S1C1, S2C2}
        hoods_to_remove = set()
        for hood, cs in embs_d[radius]["nhbds"].items():
            samples_hood = set([c.split("C")[0] for c in cs])
            if len(samples_hood) == len(cs):
                hoods_to_remove.add(hood)
        contigs_to_clear_neighs = set(
            [c for h in hoods_to_remove for c in embs_d[radius]["nhbds"][h]]
        )
        print(
            "%i hoods with %i contigs will be removed since they contain only contigs from dif samples"
            % (len(hoods_to_remove), len(contigs_to_clear_neighs))
        )

        embs_d[radius]["neighs_cleared"] = find_neighbours_optimized(
            embeddings_bincontigs,
            contignames,
            set(np.where(embeddings_mask == False)[0]).union(
                set([c_idx_d[c] for c in contigs_to_clear_neighs])
            ),
            c_cc_d,
            radius,
        )

        embs_d[radius]["nhbds_cleared"], embs_d[radius]["nhbds_cleared_g"] = (
            get_neighbourhoods(embs_d[radius]["neighs_cleared"][0], contignames)
        )

        t1 = time.time()

        print(
            "Neighbours and neighbourhoods computed in %.2f minutes" % ((t1 - t0) / 60)
        )

        if args.contignames == None:
            print(
                "Generating only hoods files, not neighbours.npz since contignames were not provided"
            )
        else:
            print("Generating hoods, neighbours and mask neighbours files")

        # neighs, _ = embs_d[radius]["neighs"]

        contigs_w_neighs = np.sum(
            [1 for row in embs_d[radius]["neighs"][0] if len(row) > 0]
        )
        print("%i contigs with neighs" % contigs_w_neighs)

        if args.contignames != None:
            neighs_file = os.path.join(
                args.neighs_outdir, "neighs_object_r_%s.npz" % (str(radius))
            )
            np.savez_compressed(
                neighs_file,
                embs_d[radius]["neighs"][0],
            )
            print(f"Neighs saved in {neighs_file}")

            ## also save neighs where only intra edges hoods are removed
            neighs_file = os.path.join(
                args.neighs_outdir,
                "neighs_intraonly_rm_object_r_%s.npz" % (str(radius)),
            )
            np.savez_compressed(
                neighs_file,
                embs_d[radius]["neighs_cleared"][0],
            )
            print(
                f"Neighs where only intra edges hoods are removed by sample saved in {neighs_file}"
            )

            ## also save neighs split by sample
            neighs_file = os.path.join(
                args.neighs_outdir, "neighs_split_object_r_%s.npz" % (str(radius))
            )
            np.savez_compressed(
                neighs_file,
                embs_d[radius]["neighs_split_by_sample"][0],
            )
            print(f"Neighs split by sample saved in {neighs_file}")

        hoods_clusters_path = os.path.join(
            args.neighs_outdir, "hoods_clusters_r_%s.tsv" % (str(radius))
        )
        with open(hoods_clusters_path, "w") as f:
            f.write("clustername\tcontigname\n")
            for hood_i, cs in embs_d[radius]["nhbds"].items():
                for c in cs:
                    f.write("%s\t%s\n" % (hood_i, c))

        hoods_wointraonly_clusters_path = os.path.join(
            args.neighs_outdir, "hoods_intraonly_rm_clusters_r_%s.tsv" % (str(radius))
        )
        ## also save cleared hoods where only intra edges hoods are removed
        hoods_wointraonly_clusters_path = os.path.join(
            args.neighs_outdir, "hoods_intraonly_rm_clusters_r_%s.tsv" % (str(radius))
        )

        with open(hoods_wointraonly_clusters_path, "w") as f:
            f.write("clustername\tcontigname\n")
            for hood_i, cs in embs_d[radius]["nhbds_cleared"].items():
                for c in cs:
                    f.write("%s\t%s\n" % (hood_i, c))

        print(
            f"Hoods where only intra edges hoods are removed by sample clusters saved in {hoods_clusters_path}"
        )

        contignames_not_in_hoods = set(contignames) - set(
            [c for cs in embs_d[radius]["nhbds"].values() for c in cs]
        )

        hoods_clusters_complete_path = os.path.join(
            args.neighs_outdir, "hoods_clusters_r_%s_complete.tsv" % (str(radius))
        )
        with open(hoods_clusters_complete_path, "w") as f:
            f.write("clustername\tcontigname\n")
            for hood_i, cs in embs_d[radius]["nhbds"].items():
                for c in cs:
                    f.write("%s\t%s\n" % (hood_i, c))
            for i, c in enumerate(contignames_not_in_hoods):
                f.write("%i_\t%s\n" % (i, c))
        print(f"Hoods clusters complete saved in {hoods_clusters_complete_path}")

    # Save the dictionary to a Pickle file
    # path_dict = os.path.join(embs_dir, "tmp/embs_d_%s.pkl"%(date))
    path_dict = os.path.join("%s/embs_d.pkl" % args.neighs_outdir)

    with open(path_dict, "wb") as pickle_file:
        pickle.dump(embs_d, pickle_file)

    print(f"Dictionary saved to {path_dict}")



# import node2vec
import argparse
import os
import pickle
import time

import networkx as nx
import numpy as np
from fastnode2vec import Graph, Node2Vec
from git_commit import get_git_commit


def remove_singleton_ccs(graph, binning_contigs):
    """
    Remove graph components if have less than 2 contigs above min_contig_size and nodes without edges

    """
    nodes_to_remove = set(
        [
            c
            for cc in nx.connected_components(graph)
            if len(cc.intersection(binning_contigs)) < 2
            for c in cc
        ]
    )
    print(
        "%i/%i contigs will be removed from the graph since they are part of components that have less than 2 binning contigs."
        % (len(nodes_to_remove), len(graph.nodes()))
    )
    print(
        "%i(%i) nodes(edges) before removing components containing less than 2 binning contig."
        % (len(graph.nodes()), len(graph.edges()))
    )
    for c in nodes_to_remove:
        graph.remove_node(c)
    print(
        "%i(%i) nodes(edges) after removing components containing less than 2 binning contig."
        % (len(graph.nodes()), len(graph.edges()))
    )

    isolated = list(nx.isolates(graph))
    graph.remove_nodes_from(isolated)

    return graph


def run_n2v(G, emb_n, wl, nw, normalization_scheme, ws, p, q):
    # Create a node2vec object
    graph = Graph(
        [(u, v, G[u][v][normalization_scheme]) for (u, v) in G.edges()],
        directed=False,
        weighted=True,
    )
    # Node2Vec(graph: fastnode2vec.graph.Graph, dim: int, walk_length: int, window: int, p: float = 1.0, q: float = 1.0, workers: int = 1, batch_walks: Optional[int] = None, use_skipgram: bool = True, seed: Optional[int] = None, **kwargs)
    n2v = Node2Vec(graph, dim=emb_n, walk_length=wl, window=ws, p=p, q=q, workers=8)

    n2v.train(epochs=nw)

    # Learn embeddings

    return n2v


if __name__ == "__main__":
    # Initialize the parser
    parser = argparse.ArgumentParser(description="Run n2v")

    # Add optional arguments with `--flags`
    parser.add_argument(
        "-G",
        dest="graph_file",
        required=True,
        type=str,
        help="Path to the graphml object that contains the graph",
    )

    parser.add_argument(
        "--ed",
        dest="emb_n",
        required=True,
        type=int,
        help="Dimension of the embeddings",
    )

    parser.add_argument(
        "--nw", dest="nw", required=True, type=int, help="Number of walks"
    )

    parser.add_argument("--ws", dest="ws", required=True, type=int, help="Window size")

    parser.add_argument(
        "--wl", dest="wl", required=True, type=int, help="Walk lenghths"
    )

    parser.add_argument("-p", dest="p", type=float, default=1.0, help="p n2v parameter")

    parser.add_argument("-q", dest="q", type=float, default=1.0, help="q n2v parameter")

    parser.add_argument(
        "--normE",
        dest="normalization_scheme",
        type=str,
        help="defines the edge weights to be used",
        default="weight",
    )

    parser.add_argument(
        "--outdirembs",
        dest="embs_dir",
        required=True,
        type=str,
        help="Path were embeddings will and embedded contigs (only contigs with neighbours are embedded)  be dumped (directory)",
    )

    parser.add_argument(
        "--contignames",
        required=True,
        type=str,
        help="File containing the contignames.",
    )

    ## Print git commit so we can debug
    commit_hash = get_git_commit(os.path.abspath(__file__))
    print(f"Git commit hash: {commit_hash}")

    # Parse the arguments
    args = parser.parse_args()
    print(args)
    # Use the accumulate function specified by either --sum or --multiply on the integers

    # Load graph

    # G = nx.read_graphml(args.graph_file)
    with open(args.graph_file, "rb") as picklefile:
        G = pickle.load(picklefile)

    ## Remove graph components that contain less than 2 binning contigs
    contignames = np.loadtxt(args.contignames, dtype=object)
    G_clean = remove_singleton_ccs(G, set(contignames))
    cc_l = list(nx.connected_components(G_clean))
    c_cc_map = {c: i for i, cc in enumerate(cc_l) for c in cc}
    cc_c_map = {i: set([c for c in cc]) for i, cc in enumerate(cc_l)}

    # define run_id
    id_run_g = (args.graph_file).split("/g_")[-1].replace(".pkl", "")

    sufix = "%s_nz_%s_en%i_wl%i_nw%i_Ws%i_p%.2f_q%.2f" % (
        id_run_g,
        args.normalization_scheme,
        args.emb_n,
        args.wl,
        args.nw,
        args.ws,
        args.p,
        args.q,
    )

    print("fastN2V run named %s" % sufix)

    t0 = time.time()

    # Identify and remove isolated nodes (nodes without edges)

    model = run_n2v(
        G,
        args.emb_n,
        args.wl,
        args.nw,
        args.normalization_scheme,
        args.ws,
        args.p,
        args.q,
    )

    t1 = time.time()
    ti = (t1 - t0) / 60

    print("fastN2V run %s finished in %.2f minutes" % (sufix, ti))

    # contignames files
    contigs_embedded = []
    embeddings = []

    for cc in nx.connected_components(G):
        for c in cc:
            embeddings.append(model.wv[c])
            contigs_embedded.append(c)

    # embs_dir_run = os.path.join(args.embs_dir, sufix + "/")
    embs_dir_run = args.embs_dir
    try:
        os.makedirs(embs_dir_run)
    except:
        pass

    embs_path = os.path.join(embs_dir_run, "embeddings.npz")

    contigsembs_path = os.path.join(embs_dir_run, "contigs_embedded.txt")

    np.savez(embs_path, np.array(embeddings))
    np.savetxt(contigsembs_path, np.array(contigs_embedded, dtype=object), fmt="%s")
    model.save(os.path.join(embs_dir_run, "word2vec.model"))

    print("Output directory: %s" % embs_dir_run)



import os
import shutil
from pathlib import Path

from loguru import logger

from PlasMAAG.command_line_runners import CliRunner, SnakemakeRunner


class EnvironmentManager:
    _src_dir = Path(__file__).parent
    _genomad_dir = _src_dir / "genomad_db"
    _genomad_db_exist = (_genomad_dir).exists()

    def install_genomad_db(self):
        logger.info(f"Installing Genomad database (~3.1 GB) to {self._genomad_dir}")
        snakemake_runner = SnakemakeRunner()
        # Download the directory in the location of the current file
        snakemake_runner.add_arguments(["--directory", self._src_dir])
        # We need to download the genomad tools first so conda needs to be passed to snakemake
        snakemake_runner.add_arguments(["--use-conda"])
        snakemake_runner.add_to_config("should_install_genomad=True")
        snakemake_runner.add_arguments(["-c", "1"])
        # Set target rule to genomad_db to create the database
        snakemake_runner.add_arguments(["download_genomad_db"])
        snakemake_runner.run()

    # def install_conda_environments(self):
    #     logger.info(f"Installing conda environments")
    #     snakemake_runner = SnakemakeRunner()
    #     snakemake_runner.add_arguments(["--use-conda", "--conda-create-envs-only"])
    #     snakemake_runner.run()

    def setup_environment(self):
        logger.info("Setting up environment")
        if not self._genomad_db_exist:
            self.install_genomad_db()

    def check_if_everything_is_setup(self):
        if True not in [self._genomad_db_exist]:
            logger.info("It seems the environment has not been setup")
            return False
        if not self._genomad_db_exist:
            raise click.UsageError(
                f"Could not find the genomad database, try running the tool with --setup_env"
            )
        return True



import argparse
import os

from Bio import SeqIO
from git_commit import get_git_commit
import gzip
import numpy as np 

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Write plasmid and non plasmid bins")

    # Add optional arguments with `--flags`
    parser.add_argument(
        "--cls_pl",
        required=True,
        type=str,
        help="Path to the plasmid clusters",
    )

    parser.add_argument(
        "--cls_nonpl",
        required=True,
        type=str,
        help="Path to the nonplasmid clusters",
    )

    parser.add_argument(
        "--contigs",
        required=True,
        type=str,
        help="Path to the fasta contigs file",
    )

    parser.add_argument(
        "--outdir",
        required=True,
        type=str,
        help="Path to the results directory",
    )

    parser.add_argument(
        "--min_nonplas_len",
        required=False,
        type=int,
        help="Min nonplasmid length",
        default=200_000,
    )

    parser.add_argument(
        "--min_plas_len",
        required=False,
        type=int,
        help="Min plasmid length",
        default=1,
    )

    
    ## Print git commit so we can debug
    commit_hash = get_git_commit(os.path.abspath(__file__))
    print("Git commit hash: " + commit_hash)

    # Parse the arguments
    args = parser.parse_args()
    print(args)

    # 
    plcl_cs_d = { cl:set() for cl,c in np.loadtxt(args.cls_pl,dtype=object,skiprows=1)}
    for cl,c in np.loadtxt(args.cls_pl,dtype=object,skiprows=1):
        plcl_cs_d[cl].add(c)
    plcl_len_d = { cl:np.sum([ int(c.split("length_")[1].split("_")[0]) for c in cs]) for cl,cs in plcl_cs_d.items() }
    c2plcl={ c:cl for cl,cs in plcl_cs_d.items() if plcl_len_d[cl] >= args.min_plas_len for c in cs}
    pl_cs = set(c2plcl.keys())
    
    
    nonplcl_cs_d = { cl:set() for cl,c in np.loadtxt(args.cls_nonpl,dtype=object,skiprows=1)}
    for cl,c in np.loadtxt(
        args.cls_nonpl,dtype=object,skiprows=1):
        nonplcl_cs_d[cl].add(c)
    
    nonplcl_len_d = { cl:np.sum([ int(c.split("length_")[1].split("_")[0]) for c in cs]) for cl,cs in nonplcl_cs_d.items() }
    c2nonplcl={ c:cl for cl,cs in nonplcl_cs_d.items() if nonplcl_len_d[cl] >= args.min_nonplas_len for c in cs}
    nonpl_cs = set(c2nonplcl.keys())
    
    os.makedirs(os.path.join(args.outdir,"candidate_plasmids"))
    os.makedirs(os.path.join(args.outdir,"candidate_genomes"))
    
    with gzip.open(args.contigs, "rt") as handle:
        for record in SeqIO.parse(handle, "fasta"):
            contig_name= record.id
            if contig_name not in nonpl_cs.union(pl_cs):
                continue
            bin_dir =  "candidate_plasmids" if contig_name in pl_cs else "candidate_genomes"
            bin_name= c2plcl[contig_name] if bin_dir == "candidate_plasmids" else c2nonplcl[contig_name]
            bin_file = bin_name+".fna"
            bin_path = os.path.join(args.outdir,"%s/%s"%(bin_dir,bin_file))
            with open(bin_path, "a") as out:
                SeqIO.write(record, out, "fasta")

        



from pathlib import Path

import rich_click as click
from loguru import logger


class OneOrMoreSnakemakeArguments(click.ParamType):
    name = "str"

    def convert(self, value, param, ctx):
        value = value.split()
        return value


class WssFile(click.ParamType):
    """
    click Type for validating whitespace separated files.
    """

    name = "White Space Separated File"

    def __init__(
        self, expected_headers, none_file_columns=[], spades_column=None
    ) -> None:
        self.expected_headers = expected_headers
        self.none_file_columns = none_file_columns
        self.spades_column = spades_column

    def convert(self, value, param, ctx):
        # Read pandas in here, to not slow cli down if pandas is not used
        try:
            import pandas as pd
        except ModuleNotFoundError as e:
            print(
                """\nCould not find module pandas, please make sure to run application in an environment containing it\n"""
            )
            raise e

        # Assert if file exist
        if not Path(value).exists():
            self.fail(f"{value!r} is not a valid path", param, ctx)

        # Assert if file is in the valid format
        try:
            df = pd.read_csv(value, sep=r"\s+")
        except pd.errors.EmptyDataError as e:
            raise self.fail(
                f"{value!r} is an empty file\nError Message:\n{str(e)}", param, ctx
            )
        except pd.errors.ParserError as e:
            raise self.fail(
                f"{value!r} is not a correctly formatted file\nError Message:\n{str(e)}",
                param,
                ctx,
            )

        # Assert if file contains the correct column headers
        if sorted(list(df.columns)) != sorted(self.expected_headers):
            raise self.fail(
                f"{value!r} Does not contain the correct headers\nError Message:\nExpected headers: {self.expected_headers}, found headers: {list(df.columns)}",
                param,
                ctx,
            )

        # Assert if file has missing values
        columns_withnan = df[df.isnull().any(axis=1)]
        if len(columns_withnan) != 0:
            raise self.fail(
                f"{value!r} is missing values or is formatted incorrectly\nError Message:\nThe following columns have missing values:\n{columns_withnan.to_markdown(index=False)}",
                param,
                ctx,
            )

        # Assert if all paths to files in the file exist
        # TODO make is such that it throws errors for all files which do not exist instead of 1 at a time
        file_exist = lambda file: Path(str(file)).exists()
        parts_of_df_which_are_files = df.drop(columns=self.none_file_columns)
        for i, content in enumerate(
            parts_of_df_which_are_files.itertuples(index=False)
        ):
            for file in list(content):
                if not file_exist(file):
                    raise self.fail(
                        f"{value!r} has file(s) which do not exist\nError Message:\n < {value} > Contains column < {i +1 } > (1 indexed) which has the file < {file} > which do not exist",
                        param,
                        ctx,
                    )

        # Assert if spades columns are correct
        if self.spades_column is not None:
            [
                self.is_spades_dir_correct(x, value, param, ctx)
                for x in df[self.spades_column]
            ]

        logger.info(
            f"\nRead in the following sample list from '{value}' using flag '--{param.human_readable_name}':\n {df.to_markdown(index=False)}"
        )

        print("")
        return value

    def is_spades_dir_correct(self, x, value, param, ctx):
        if not Path(x).is_dir():
            raise self.fail(
                f"{value!r} has defined a Spades directory which is not a directory\nError Message:\n < {value} > contains '{x}' which is not a dir",
                param,
                ctx,
            )
        for file in [
            "contigs.fasta",
            "assembly_graph_after_simplification.gfa",
            "contigs.paths",
        ]:
            if not (Path(x) / file).exists():
                raise self.fail(
                    f"{value!r} has defined a Spades directory ('{x}') which does not contain {file}\nError Message:\n < {value} > contains '{x}' which does not contain '{file}'. Are you sure '{x}' is a spades output directory?",
                    param,
                    ctx,
                )
